!  rspmi.f90
!  
!  Copyright 2024 Igor J. Chybicki <igorchy@ukw.edu.pl>
!  
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
!  
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!  
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
!  MA 02110-1301, USA.
!  
!  Info:
!  rspmi estimates seed and pollen migration rates based on
!  a sample of individuals from a closed network of populations
!  The program generates estimates under assumption that migration
!  rates are low, with the total immigration rate = l+(1-l)z, were
!  l and z is a fixed (a priori) probability of non-immigration, and
!  the isolation parameter l,z=(0,1).

program rspmi
   use omp_lib
   
   implicit none
   
   !----------------------------------------------------------------------------------------------------------
   ! global declarations
   !
   ! constants
   character(1), parameter :: tab=char(9), creturn = achar(13) 
   double precision, parameter :: mypi = dacos(dble(-1.0)), my2pi = 2.D0 * mypi
   double precision, parameter :: MINQLM = 0.001, MAXQLM = 0.999  ! min/max limits for global allele frequencies
   double precision, parameter :: MINFK = 0.001, MAXFK = 0.999    ! min/max limits for divergence rates (population Fst)
   double precision, parameter :: MINYR = 0.001, MAXYR = 0.999    ! min/max limits for dispersion parameter (inbreeding)
   double precision, parameter :: MINYA = 0.001, MAXYA = 0.999    ! min/max limits for dispersion parameter (alpha)
   double precision, parameter :: MINYB = 0.001, MAXYB = 0.999    ! min/max limits for dispersion parameter (beta)
   double precision, parameter :: MINZA = 0.001, MAXZA = 0.999    ! min/max limits for isolation parameter (alpha)
   double precision, parameter :: MINZB = 0.001, MAXZB = 0.999    ! min/max limits for isolation parameter (beta)
   double precision, parameter :: MINBA = -19.999, MAXBA = 19.999 ! min/max limits for regression slope (alpha)
   double precision, parameter :: MINBB = -19.999, MAXBB = 19.999 ! min/max limits for regression slope (beta)
   
   ! useful variables
   integer, dimension(8) :: tvalues                     ! date-time variables used to initialize RNG, if needed
   character(255) :: input_file, output_file, data_file ! file names
   character(len=255), allocatable :: pname(:)          ! parameter names
   integer :: nthreads                                  ! number of parallel threads
   
   ! RNG-related variables
   integer*8 :: seed,jsr              ! serial version
   integer*8, allocatable :: pjsr(:)  ! parallel version
   
   ! data dimensions
   integer :: npop, nind, nloc, maxnall         ! number of pops, individuals, loci, alleles
   integer, allocatable :: nall(:), allele(:,:) ! number of alleles at individual loci, allele codes
   
   ! mcmc settings
   integer :: ns, nb, ke, nu    ! number of sampling and burn-in iterations, thinning, number of saved updates
   logical :: inbreeding        ! if TRUE: inbreeding is added to the model
   logical :: dropout           ! if TRUE: dropout is added to the model
   logical :: zygotic_migration ! if TRUE: zygotic (seed) migration is added to the model
   logical :: gametic_migration ! if TRUE: whether gametic (pollen) is added to the model
   logical :: coordinates       ! if TRUE: coordinates are available
   logical :: planexy           ! if TRUE: coordinates are given as Carthesian XY
   logical :: spatial_zygotic   ! if TRUE: isolation model for zygotic migration is spatially-explicit
   logical :: spatial_gametic   ! if TRUE: isolation model for gametic migration is spatially-explicit
   logical :: rjmcmc            ! if TRUE: the probability of spatially-explicit isolation models is assessed
   logical :: ideal_data        ! if TRUE: data are single-locus pseudo-genotypes with alleles denoting seed and pollen ancestry
   logical :: optimize_threads  ! if TRUE: number of parallel threads is optimized to get the best overall performance
   logical :: convert_data      ! if TRUE: data file is converted into BIMr and BayesAss input files (saved text files)
   
   ! hyper-parameters
   ! these parameters are related directly to para meters, while indirectly to data
   double precision :: y_alpha, z_alpha, b_alpha    ! hyper-parameters of the prior for alpha
   double precision :: y_beta, z_beta, b_beta       ! hyper-parameters of the prior for beta
   double precision :: m_f, y_f, m_r, y_r           ! hyper=parameters of the prior for divergence (fst) and inbreeding (fis)
   double precision, allocatable :: qlm(:,:), fk(:) ! global allele frequencies, population divergence rates
   double precision :: v_b = 100.D0   ! the variance of prior normal distribution for distance effect (regression slope)
   double precision :: l_alpha = 0.D0 ! prior zygotic non-imigration (if zero, z_alpha controls non-imigration entirely)
   double precision :: l_beta = 0.D0  ! prior gametic non-imigration (if zero, z_beta controls non-imigration entirely)
   double precision, allocatable :: r_alpha(:,:), r_beta(:,:) ! spatial model expectations for alpha and beta
   double precision :: dist_limit = 0.D0 !distance limit used to make dispersal kernel a proper density function (neglected if zero)
   
   ! parameters (allele frequencies in pops, gametic and zygotic migration matrices, dropout and inbreeding rates)
   ! these parameters are related directly to data (genotype structure)
   double precision, allocatable :: pklm(:,:,:), alpha(:,:), beta(:,:), dl(:), ri(:), mjk(:,:)
   
   ! parameter means (m_), variances (v_), and thinned lists across mcmc (l_) (where _ is a variable name)
   double precision :: mlogPrAugData, vlogPrAugData,mPrSpatZyg,mPrSpatGam
   double precision, allocatable :: mpklm(:,:,:), malpha(:,:), mbeta(:,:), mdl(:), mri(:), mmjk(:,:)
   double precision, allocatable :: vpklm(:,:,:), valpha(:,:), vbeta(:,:), vdl(:), vri(:), vmjk(:,:)
   double precision, allocatable :: lalpha(:,:,:), lbeta(:,:,:), ldl(:,:), lri(:,:)
   double precision :: my_alpha,my_beta, mz_alpha, mz_beta, mm_f, my_f, mm_r, my_r, mb_alpha, mb_beta
   double precision :: vy_alpha,vy_beta, vz_alpha, vz_beta, vm_f, vy_f, vm_r, vy_r, vb_alpha, vb_beta
   double precision, allocatable :: ly_alpha(:),ly_beta(:), lz_alpha(:), lz_beta(:), lm_r(:), ly_r(:), lb_alpha(:), lb_beta(:)
   double precision, allocatable :: mqlm(:,:), mfk(:)
   double precision, allocatable :: vqlm(:,:), vfk(:)
   double precision, allocatable :: lm_f(:),lfk(:,:)
   double precision, allocatable :: lmjk(:,:,:)
   integer, allocatable :: n_ianc(:,:,:) ! visiting counts across a thinned list of invididual ancestries
   
   ! model support (WAIC)
   double precision, allocatable :: iprob(:), liprob(:,:) ! point-wise probability, and thinned list (l_)
   double precision :: lppd, pw, s_waic
   
   ! RJ-MCMC proposal distributions for spatial parameters (regression slopes)
   double precision :: mq_b_alpha,vq_b_alpha,mq_b_beta,vq_b_beta ! m-mean, v-variance of normal distrib.
   
   ! binary indicators for spatial models (incl = 1 denotes spatial model to be switched on)
   integer :: incl_b_alpha,incl_b_beta
   
   ! number of mcmc samples for spatial parameters taken under the spatial model
   integer ::nu_b_alpha,nu_b_beta
   
   ! proposal distribution-related variables
   double precision :: accept_q,reject_q,delta_q
   double precision :: accept_f,reject_f,delta_f
   double precision :: accept_mf,reject_mf,delta_mf
   double precision :: accept_md,reject_md,delta_md
   double precision :: accept_mr,reject_mr,delta_mr
   double precision :: accept_yr,reject_yr,delta_yr
   double precision :: accept_ya,reject_ya,delta_ya
   double precision :: accept_yb,reject_yb,delta_yb
   double precision :: accept_za,reject_za,delta_za
   double precision :: accept_zb,reject_zb,delta_zb
   double precision :: accept_ba,reject_ba,sigma_ba
   double precision :: accept_bb,reject_bb,sigma_bb
   
   ! data provided by user:
   ! ipop - invidual population indicator
   ! igen - individual genotype
   ! inam - individual indicator
   ! pop - population indicator
   ! nindpop - number of individuals per pop
   ! ilon,ilat - individual coordinates
   ! plon,plat - population coordinates
   ! pdist - square matrix of pairwise distances between populations
   ! rlm - observed mean allele frequencies
   ! rklm - observed population (sampling location) allele frequencies
   integer, allocatable :: ipop(:), igen(:,:,:), inam(:), pop(:), nindpop(:)
   double precision, allocatable :: ilon(:), ilat(:), plon(:), plat(:), pdist(:,:), rlm(:,:), rklm(:,:,:)
   
   ! missing data (auxiliary variables) used in the Gibbs sampler:
   ! ianc - individual ancestry
   ! igenanc - individual gene ancestry
   ! idout - individual scores of allele dropout
   ! nak - number of individuals in sampling locations with seeds taken from a given populations
   ! nbk - number of individuals in sampling locations with pollen taken from a given populations
   ! nklm - number of allele copies in source populations (before migration)
   ! ndl - number of allelic dropout events per locus
   ! iinb - number of loci with ibd homozygosity per individual (for inbreeding)
   integer, allocatable :: ianc(:,:),igenanc(:,:,:),idout(:,:), nak(:,:), nbk(:,:), nklm(:,:,:), ndl(:,:),iinb(:,:)
   
   !----------------------------------------------------------------------------------------------------------
   
   !----------------------------------------------------------------------------------------------------------
   ! PROGRAM STARTS HERE
   
   ! get data file name as a command-line argument (required)
   call get_command_argument(1, input_file)
   if (len_trim(input_file) == 0) then
     write (*,"('File name not specified! (a command-line agrument is missing, e.g. ""./rspmi my_project.txt"")')")
     call exit
   endif
   
   ! read options and data
   call read_input_file
   
   ! number of parallel threads
   if (nthreads==0) then
     optimize_threads=.true.
     nthreads=omp_get_max_threads()
   endif
      
   ! if seed=0, rng seeded with the time (in milliseconds) since midnight
   if (seed==0) then
     call date_and_time(values=tvalues)
     seed=((tvalues(5)*60+tvalues(6))*60+tvalues(7))*1000+tvalues(8)
   endif
   jsr=seed
   call init_rnd
   call init_prnd
   
   ! drop-out (used for simulations; if the first parameter is zero then data remains unaffected)
   call input_dropout(0.D0,0.1D0,0.75D0)
   
   ! run MCMC sampler
   call run_mcmc
   
   ! save BA3 & BIMr data
   if (convert_data) then
     call BA3_data_file
     call BIMr_data_file
   endif
   
   ! PROGRAM ENDS HERE
   !----------------------------------------------------------------------------------------------------------
   
 contains
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine modifies data to get heterozygous genotypes after random allelic drop-out
 ! useful for simulation study
 subroutine input_dropout(pdout,edout,maxdout)
   integer :: i,m,l
   double precision :: pdout,edout,maxdout,dout(nloc)
   character(len=255) :: filename,extension
   
   if (edout<=0.D0 .or. edout>=1.D0 .or. pdout<=0.D0) return
   
   do
     do l=1,nloc
       dout(l)=0.D0
       if (rnd()<pdout) then
         do
           dout(l)=-dlog(1.D0-rnd())*edout
           if (dout(l)>0.D0 .and. dout(l)<maxdout) exit
         enddo
       endif
     enddo
     
     if (sum(dout)>0.D0) exit
   enddo
   
   do l=1,nloc
     do i=1,nind
       if (igen(i,l,1)>0 .and. igen(i,l,2)>0 .and. igen(i,l,1)/=igen(i,l,2)) then
         if (rnd()<dout(l)) then
           m=int(rnd()*2)+1
           igen(i,l,m)=igen(i,l,3-m)
         endif
       endif
     enddo
   enddo
   
   extension='edout'
   filename=output_filename(extension)
   open (unit=789,file=trim(filename),action="write",status="replace")
   write (789,"('Expected drop-out:')")
   write (789,"('marker',a,'dout')") tab
   do l=1,nloc
     write (789,"('loc_',i0,a,f7.4)") l,tab,dout(l)
   enddo
   close(789)
   write (*,"(a,f7.4)") '[rspmi] Random drop-out simulated at frequency: ',sum(dout)/dble(nloc)
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine verifies probability for single-locus genotype (whether all terms sum up to 1)
 subroutine test_pr_gil
   integer :: i,j,k,a,b,l,p
   double precision :: pr,s1,s2
   
   return ! comment out this line to run the test
   
   i=1
   l=1
   p=1
   s1=0.D0
   do j=1,npop
     do k=1,npop
       s2=0.D0
       do a=1,nall(l)
         do b=a,nall(l)
           igen(i,l,1)=a
           igen(i,l,2)=b
           pr=pr_gil(1,l,j,k)
           s1=s1+alpha(p,j)*beta(j,k)*pr
           s2=s2+pr
         enddo
       enddo
       write (*,*) i,j,k,l,s2
     enddo
   enddo
   write (*,*) 'total ',s1,s2
   read (*,*)
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine creates output file name with specified extension based on the name of input file
 function output_filename(extension)
   character(len=255) :: extension,output_filename,ofn
   integer :: i,j
   
   j=0
   ofn='rspmi.'//trim(extension)
   do i=1,len(trim(input_file))
     if (input_file(i:i)=='.' .and. .not. input_file(i+1:i+1)=='/') then
       j=i
       write(ofn,"(a,'.',a)") input_file(1:i-1),trim(extension)
       exit
     endif
   enddo
   if (j==0) write(ofn,"(a,'.',a)") input_file(1:len(trim(input_file))),trim(extension)
   output_filename=ofn
 end function
 !----------------------------------------------------------------------------------------------------------
  
 !----------------------------------------------------------------------------------------------------------
 ! This routine specifies list of parameter names used in the output file
 subroutine parameter_names
   integer :: i,j,k,nparam
   
   nparam=npop+1
   if (zygotic_migration) nparam=nparam+npop**2+2
   if (zygotic_migration .and. spatial_zygotic) nparam=nparam+1
   if (gametic_migration) nparam=nparam+npop**2+2
   if (gametic_migration .and. spatial_gametic) nparam=nparam+1
   if (inbreeding) nparam=nparam+nind+2
   if (dropout) nparam=nparam+nloc
   
   allocate(pname(nparam))
   
   k=0
   
   if (zygotic_migration) then
     do i=1,npop
       do j=1,npop
         k=k+1
         write(pname(k),"('alpha[',i0,',',i0,']')") i,j
       enddo
     enddo
     
     k=k+1
     write(pname(k),"('y_alpha')")
     k=k+1
     write(pname(k),"('z_alpha')")
     
     if (zygotic_migration .and. spatial_zygotic) then
       k=k+1
       write(pname(k),"('b_alpha')")
     endif
   endif
   
   if (gametic_migration) then
     do i=1,npop
       do j=1,npop
         k=k+1
         write(pname(k),"('beta[',i0,',',i0,']')") i,j
       enddo
     enddo
     
     k=k+1
     write(pname(k),"('y_beta')")
     k=k+1
     write(pname(k),"('z_beta')")
     
     if (gametic_migration .and. spatial_gametic) then
       k=k+1
       write(pname(k),"('b_beta')")
     endif
   endif
   
   if (dropout) then
     do i=1,nloc
       k=k+1
       write(pname(k),"('dout[',i0,']')") i
     enddo
   endif
   
   if (inbreeding) then
     do i=1,nind
       k=k+1
       write(pname(k),"('fis[',i0,']')") i
     enddo
     
     k=k+1
     write(pname(k),"('m_fis')")
     k=k+1
     write(pname(k),"('y_fis')")
   endif
   
   do i=1,npop
     k=k+1
     write(pname(k),"('fst[',i0,']')") i
   enddo
   
   k=k+1
   write(pname(k),"('m_fst')")
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine is a simple but effective way to lower case of letter characters
 subroutine lower_case_string(string)
   character(len=5) :: string
   character(len=26) :: lower_case = 'abcdefghijklmnopqrstuvwxyz'
   character(len=26) :: upper_case = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
   integer :: i,j
   
   do i=1,5
     j = index(upper_case,string(i:i))
     if (j /= 0) string(i:i) = lower_case(j:j) 
   enddo
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine reads input file
 subroutine read_input_file
   integer :: i,l,m
   character(len=255) :: line,option
   logical :: read_data,options_defined
   
   ! here are default settings:
   seed=0
   ns=20000
   nb=10000
   ke=10
   nthreads=0
   nind=0
   nloc=0
   l_alpha=2.D0/3.D0
   l_beta=2.D0/3.D0
   dropout=.false.
   inbreeding=.false.
   zygotic_migration=.false.
   gametic_migration=.false.
   coordinates=.false.
   zygotic_migration=.false.
   zygotic_migration=.false.
   spatial_zygotic=.false.
   spatial_gametic=.false.
   planexy=.false.
   rjmcmc=.false.
   ideal_data=.false.
   options_defined=.false.
   optimize_threads=.false.
   convert_data=.false.
   data_file="none"
   
   ! check if the file exists
   if (access(trim(input_file),'r') .ne. 0) then
     write (*,"(a)") '[rspmi] File "'//trim(input_file)//'" does not seem to exist. Press <ENTER> to terminate...'
     read (*,*)
     call exit
   endif
   
   open (unit=789,file=trim(input_file),action="read")
   read_data = .false.
   ! read data dimensions, mcmc & model settings, and the number of parallel threads
   do
     read (789,"(a)") line
     write(option,"(a)") trim(adjustl(line))
     call lower_case_string(option(1:5)) ! this makes the options be case-insensitive (first 5 characters)
     select case (option(1:5))
       case ('#samp')
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) ns
             write (*,"(a,i0)") '[rspmi] User-defined length of MCMC sampling stage: ',ns
             exit
           endif
         enddo
       case ('#burn')
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) nb
             write (*,"(a,i0)") '[rspmi] User-defined length of MCMC burn-in stage: ',nb
             exit
           endif
         enddo
       case ('#keep')
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) ke
             write (*,"(a,i0)") '[rspmi] User-defined thinning parameter (keep-every): ',ke
             exit
           endif
         enddo
       case ('#indi')
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) nind
             write (*,"(a,i0)") '[rspmi] User-defined number of individuals: ',nind
             exit
           endif
         enddo
       case ('#mark')
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) nloc
             write (*,"(a,i0)") '[rspmi] User-defined number of markers: ',nloc
             exit
           endif
         enddo
       case ('#thre')
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) nthreads
             if (nthreads>0) then
               write (*,"(a,i0)") '[rspmi] User-defined number of parallel threads: ',nthreads
               exit
             else
               write (*,"(a)") '[rspmi] User-defined number of parallel threads: optimized'
               exit
             endif
           endif
         enddo
       case ('#seed')
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) seed
             if (seed>0) then
               write (*,"(a,i0)") '[rspmi] User-defined RNG seed: ',seed
               exit
             else
               write (*,"(a)") '[rspmi] User-defined RNG seed: based on the system clock'
               exit
             endif
           endif
         enddo
       case ('+spat')
         do i=6,len(trim(option))
           if (option(i:i)=='_') then
             call lower_case_string(option(i+1:i+5))
             if (option(i+1:i+5)=='zygot') then
               write (*,"(a,i0)") '[rspmi] User-defined zygotic isolation model: spatially-explicit'
               spatial_zygotic=.true.
               exit
             else if (option(i+1:i+5)=='gamet') then
               write (*,"(a,i0)") '[rspmi] User-defined gametic isolation model: spatially-explicit'
               spatial_gametic=.true.
               exit
             endif
           endif
         enddo
       case ('+idea')
         ideal_data=.true.
         write (*,"(a,i0)") '[rspmi] User-defined data model: ideal ancestry'
       case ('+rjmc')
         rjmcmc=.true.
       case ('#coor')
         coordinates=.true.
         write (*,"(a,i0)") '[rspmi] Geographic coordinates: provided as lon/lat'
       case ('#plan')
         coordinates=.true.
         planexy=.true.
         write (*,"(a,i0)") '[rspmi] Geographic coordinates: provided as x/y'
       case ('+drop')
         dropout=.true.
         write (*,"(a,i0)") '[rspmi] User-defined genotyping model: allelic drop-out allowed'
       case ('+inbr')
         inbreeding=.true.
         write (*,"(a,i0)") '[rspmi] User-defined mating model: inbreeding allowed'
       case ('+dist')
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) dist_limit
             write (*,"(a,i0)") '[rspmi] User-defined dispersal distance limit: '
             exit
           endif
         enddo
       case ('+zygo')
         zygotic_migration=.true.
         write (*,"(a,i0)") '[rspmi] User-defined migration model: zygotic migration allowed'
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) l_alpha
             if (l_alpha<1.D0) then
               write (*,"(a,f6.4)") '[rspmi] User-defined prior zygotic isolation: ',l_alpha
               exit
             endif
           endif
         enddo
         if (l_alpha==1.D0) then
           write (*,"(a,f6.4)") '[rspmi] User-defined prior zygotic isolation: complete (zygotic migration disabled)'
           zygotic_migration=.false.
         endif
       case ('+game')
         gametic_migration=.true.
         write (*,"(a,i0)") '[rspmi] User-defined migration model: gametic migration allowed'
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) l_beta
             if (l_beta<1.D0) then
               write (*,"(a,f6.4)") '[rspmi] User-defined prior gametic isolation: ',l_beta
               exit
             endif
           endif
         enddo
         if (l_beta==1.D0) then
           write (*,"(a,f6.4)") '[rspmi] User-defined prior gametic isolation: complete (gametic migration disabled)'
           gametic_migration=.false.
         endif
       case ('#data')
         do i=6,len(trim(option))
           if (option(i:i)=='=') then
             read (option(i+1:len(trim(option))),*) data_file
             if (access(trim(data_file),'r') .ne. 0) then
               write (*,"(a)") '[rspmi] Data file "'//trim(input_file)//'" does not seem to exist. Exiting...'
               stop
             else
               write (*,"(2a)") '[rspmi] User-defined data file: ',trim(data_file)
               read_data=.true.
               exit
             endif             
           endif
         enddo
       case ('+conv')
         convert_data=.true.
         write (*,"(a,i0)") '[rspmi] Input file conversion requested'
       case ('#end')
         options_defined=.true.        
       case default
         write (*,"(a)") '[rspmi] Failed to read options in the input file. Exiting...'
         stop
     end select
     if (options_defined) exit
   enddo
   
   if (.not. inbreeding) write (*,"(a,i0)") '[rspmi] User-defined mating model: inbreeding NOT allowed'
   if (.not. dropout) write (*,"(a,i0)") '[rspmi] User-defined genotyping model: allelic drop-out NOT allowed'
   if (.not. zygotic_migration) write (*,"(a,i0)") '[rspmi] User-defined migration model: zygotic migration NOT allowed'
   if (.not. gametic_migration) write (*,"(a,i0)") '[rspmi] User-defined migration model: gametic migration NOT allowed'
   if (.not. coordinates) then
     write (*,"(a,i0)") '[rspmi] Geographic coordinates: NOT provided'
     spatial_gametic=.false.
     spatial_zygotic=.false.
   endif
   if (.not. coordinates) rjmcmc=.false.
   if (.not. spatial_zygotic) write (*,"(a,i0)") '[rspmi] User-defined zygotic isolation model: NON-spatial'
   if (.not. spatial_gametic) write (*,"(a,i0)") '[rspmi] User-defined gametic isolation model: NON-spatial'
   if (.not. spatial_zygotic .and. .not. spatial_gametic) rjmcmc=.false.
   if (rjmcmc) then
     write (*,"(a,i0)") '[rspmi] RJ-MCMC for spatial component(s): enabled'
   else
     write (*,"(a,i0)") '[rspmi] RJ-MCMC for spatial component(s): disabled'
   endif
   
   if (nind==0) then
     write (*,"(a)") '[rspmi] Number of individuals must be specified. Exiting...'
   endif
   
   if (nloc==0) then
     write (*,"(a)") '[rspmi] Number of markers must be specified. Exiting...'
   endif
   close (789)
   flush(6)
   
   if (read_data) then
     open (unit=789,file=trim(data_file),action="read")
     allocate (igen(nind,nloc,2))
     allocate (ipop(nind))
     allocate (inam(nind))
     if (coordinates) allocate(ilon(nind))
     if (coordinates) allocate(ilat(nind))
     if (coordinates) then
       do i=1,nind
         read (789,*) inam(i),ipop(i),ilon(i),ilat(i),((igen(i,l,m),m=1,2),l=1,nloc)
       enddo
     else
       do i=1,nind
         read (789,*) inam(i),ipop(i),((igen(i,l,m),m=1,2),l=1,nloc)
       enddo
     endif
     close (789)
     
     call identify_pops
     call identify_alleles
     call compute_obs_allfreqs
     call parameter_names
     if (coordinates) then 
       call compute_pop_geoloc
     else
       allocate(pdist(npop,npop))
       pdist=1.D0
     endif
   close (789)
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine identifies pops
 subroutine identify_pops
   integer :: i,j,p
   integer, allocatable :: tmppop(:)
   
   ! identify all pop identifiers
   ! and keep them in 'pop' array
   npop=0
   allocate(pop(npop))
   do i=1,nind
     p=0
     do j=1,npop
       if (ipop(i)==pop(j)) p=1
     enddo
     
     if (p==0) then
       allocate(tmppop(npop))
       tmppop=pop
       deallocate(pop)
       npop=npop+1
       allocate(pop(npop))
       pop(1:npop-1)=tmppop
       pop(npop)=ipop(i)
       deallocate(tmppop)
     endif
   enddo
   
   ! individual pop identifiers coded
   ! with integer indices in the pop array
   if (ideal_data) then
     allocate(ianc(nind,2))
     ianc(1:nind,1:2)=igen(1:nind,1,1:2)
   endif
   
   ! individual pop identifiers coded
   ! with integer indices in the pop array
   do i=1,nind
     do j=1,npop
       if (ipop(i)==pop(j)) then
         ipop(i)=j
         exit
       endif
     enddo
   enddo
   
   ! count individuals in pops
   allocate(nindpop(npop))
   nindpop=0
   do i=1,nind
     nindpop(ipop(i))=nindpop(ipop(i))+1
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! identifies alleles
 subroutine identify_alleles
   ! double '-1' used for missing genotype
   integer, allocatable ::  tmpallele(:,:)
   integer :: i,j,k,l,m,b
 
   allocate(allele(nloc,0))
   allocate(nall(nloc))
   allocate(tmpallele(nloc,0))
   
   do i=1,nloc
     nall(i)=0
   enddo
   
   maxnall=0
   
   !alleles in individual genotypes
   do i=1,nind
     do j=1,nloc
       do k=1,2
         b=0
         do l=1,nall(j)
           if (igen(i,j,k)==allele(j,l)) then
             b=b+1
           endif
         enddo
         
         if ((igen(i,j,k).ne.-1).and.(b==0)) then
           tmpallele=allele
           nall(j)=nall(j)+1
           if (maxnall<nall(j)) then
             maxnall=nall(j)
             deallocate(allele)
             allocate(allele(nloc,maxnall))
           endif
           do l=1,nloc
             do m=1,nall(l)
               if (l.ne.j) then
                 allele(l,m)=tmpallele(l,m)
               else if (m.lt.nall(l)) then
                 allele(l,m)=tmpallele(l,m)
               endif
             enddo
           enddo
           allele(j,nall(j))=igen(i,j,k)
           deallocate(tmpallele)
           allocate(tmpallele(nloc,maxnall))
         endif
       enddo
     enddo
   enddo
   
   ! encodes alleles using their indices in the allele table
   do i=1,nind
     do j=1,nloc
       do k=1,2
         do m=1,nall(j)
           if (igen(i,j,k)==allele(j,m)) then
             igen(i,j,k)=m
             exit
           endif
         enddo
       enddo
     enddo
   enddo
     
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes central (median) geographic coordinates for pops
 subroutine compute_pop_geoloc
   integer :: i,j,k,median_i
   double precision :: mean_d,median_d
   
   allocate(plon(npop))
   allocate(plat(npop))
   
   do k=1,npop
     median_i=0
     median_d=huge(median_d)
     do i=1,nind
       if (ipop(i)/=k) cycle
       
       ! compute mean distance between i-th individual and individuals within k-th pop
       mean_d=0.D0
       do j=1,nind
         if (i==j .or. ipop(j)/=k) cycle
         mean_d=mean_d+distance(ilon(i),ilat(i),ilon(j),ilat(j))/dble(nindpop(k)-1)
       enddo
       if (median_i==0 .or. mean_d<median_d) then
         median_d=mean_d
         median_i=i
       endif
     enddo

     plon(k)=ilon(median_i)
     plat(k)=ilat(median_i)
   enddo
   
   allocate(pdist(npop,npop))
   do j=1,npop
     pdist(j,j)=0.D0
     do k=j+1,npop
       pdist(j,k)=distance(plon(k),plat(k),plon(j),plat(j))
       pdist(k,j)=pdist(j,k)
     enddo
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes pop allele frequencies based on ipop identifiers
 subroutine compute_obs_allfreqs
   integer :: i,l,m,nk(npop,nloc),n(nloc)
   
   allocate(rklm(npop,nloc,maxnall))
   allocate(rlm(nloc,maxnall))
   rlm=0.D0
   
   n=0
   nk=0
   do i=1,nind
     do l=1,nloc
       do m=1,2
         if (igen(i,l,m) /= -1) then
           rlm(l,igen(i,l,m))=rlm(l,igen(i,l,m))+1.D0
           n(l)=n(l)+1
           
           rklm(ipop(i),l,igen(i,l,m))=rklm(ipop(i),l,igen(i,l,m))+1.D0
           nk(ipop(i),l)=nk(ipop(i),l)+1
         endif
       enddo
     enddo
   enddo
   
   do l=1,nloc
     rlm(l,:)=rlm(l,:)/dble(n(l))
     
     do i=1,npop
       rklm(i,l,:)=rklm(i,l,:)+rlm(l,:)
       nk(i,l)=nk(i,l)+1
       rklm(i,l,:)=rklm(i,l,:)/dble(nk(i,l))
     enddo
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns logGamma (defined in terms of mean and shape)
 double precision function log_GammaD(x,m,k)
   double precision, intent(in) :: x,m,k
   double precision :: a,b
   
   a=k
   b=k/m
   log_GammaD=a*dlog(b)-DLGAMA(a)+(a-dble(1))*dlog(x)-b*x
   
 end function log_GammaD
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns log density of Dirichlet distribution
 double precision function log_DirichletD(x,b,n)
   integer, intent(in) :: n
   double precision, intent(in) :: x(n),b(n)
   integer :: i
   double precision :: pg,sb,p
   
   pg=dble(0)
   sb=dble(0)
   p=dble(0)
   do i=1,n
     sb=sb+b(i)
     pg=pg+DLGAMA(b(i))
     p=p+(b(i)-dble(1))*dlog(x(i))
   enddo
   log_DirichletD=p-pg+DLGAMA(sb)
   
 end function log_DirichletD
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns log density of BetaBinomial distribution
 double precision function log_BetaBinD(n1,n2,m,y)
   integer, intent(in) :: n1,n2
   double precision, intent(in) :: m,y
   double precision :: a,b
   
   a = m * (dble(1) - y) / y
   b = (dble(1) - m) * (dble(1) - y) / y
   log_BetaBinD = DLGAMA(dble(n1)+a)+DLGAMA(dble(n2)+b)-DLGAMA(dble(n1+n2)+a+b)+DLGAMA(a+b)-DLGAMA(a)-DLGAMA(b)
   
 end function log_BetaBinD
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns log density of DirichletMultinomial distribution
 double precision function log_DirichletMultD(x,u,y,n)
   integer :: n,x(n),i,sx
   double precision :: u(n),y,fy,ai
   
   fy=(1-y)/y
   sx=sum(x)
   log_DirichletMultD=DLGAMA(dble(sx+1))+DLGAMA(fy)-DLGAMA(dble(sx)+fy)
   do i=1,n
     ai=u(i)*fy
     log_DirichletMultD=log_DirichletMultD+DLGAMA(dble(x(i))+ai)-DLGAMA(dble(x(i)+1))-DLGAMA(ai)
   enddo
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns log density of Normal distribution (arguments: m-mean, v-variance)
 double precision function log_NormD(x,m,v)
   double precision, intent(in) :: x,m,v
   
   log_NormD = -(x-m)**2/(dble(2)*v)-dlog(my2pi*v)*dble(0.5)
 end function log_NormD
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns single-locus probability
 ! alleles in a homozygote can be ibd only if they are sampled from the same pop (i.e. j=k)
 function pr_gil(i,l,j,k)
   double precision :: pr_gil
   integer :: i,l,j,k
   
   pr_gil=0.D0
   if (igen(i,l,1) /= -1 .and. igen(i,l,2) /= -1) then ! missing genotypes are skipped (no information)
     if (igen(i,l,1)==igen(i,l,2)) then
       ! observed as homozygote
       if (j==k) then
         ! alleles from the same locality
         pr_gil=(1.D0-ri(i))*pklm(j,l,igen(i,l,1))**2                                         ! true homozygote non-ibd
         pr_gil=pr_gil+ri(i)*pklm(j,l,igen(i,l,1))                                            ! true homozygote ibd
         pr_gil=pr_gil+(1.D0-ri(i))*dl(l)*pklm(j,l,igen(i,l,1))*(1.D0-pklm(j,l,igen(i,l,1)))  ! true heterozygote
       else
         ! alleles from different localities
         pr_gil=pklm(j,l,igen(i,l,1))*pklm(k,l,igen(i,l,1))                           ! true homozygote
         pr_gil=pr_gil+0.5D0*dl(l)*pklm(j,l,igen(i,l,1))*(1.D0-pklm(k,l,igen(i,l,1))) ! true heterozygote 1
         pr_gil=pr_gil+0.5D0*dl(l)*pklm(k,l,igen(i,l,1))*(1.D0-pklm(j,l,igen(i,l,1))) ! true heterozygote 2
       endif
     else
       ! observed as heterozygote
       if (j==k) then
         ! alleles from the same locality
         pr_gil=(1.D0-ri(i))*(1.D0-dl(l))*2.D0*pklm(j,l,igen(i,l,1))*pklm(j,l,igen(i,l,2))
       else
         ! alleles from different localities
         pr_gil=(1.D0-dl(l))*pklm(j,l,igen(i,l,1))*pklm(k,l,igen(i,l,2))
         pr_gil=pr_gil+(1.D0-dl(l))*pklm(k,l,igen(i,l,1))*pklm(j,l,igen(i,l,2))
       endif
     endif
   else
     pr_gil=1.D0
   endif
   
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns log-probability of individual multi-locus genotype
 function log_pr_gi(i,j,k)
   double precision :: log_pr_gi
   integer :: i,j,k,l
   
   log_pr_gi=0.D0
   do l=1,nloc
     log_pr_gi=log_pr_gi+dlog(pr_gil(i,l,j,k))
   enddo
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! returns log-probability of ind
 function log_pr_ind(i)
   double precision :: log_pr_ind
   integer :: i,j,k
   
   log_pr_ind=0.D0
   do j=1,npop
     do k=1,npop
       log_pr_ind=log_pr_ind+dlog(alpha(ipop(i),j))+dlog(beta(j,k))+log_pr_gi(i,j,k)
     enddo
   enddo
   
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns log-probability of data
 function log_pr_data()
   double precision :: log_pr_data
   integer :: i,j,k
   
   log_pr_data=0.D0
   do i=1,nind
     do j=1,npop
       do k=1,npop
         log_pr_data=log_pr_data+dlog(alpha(ipop(i),j))+dlog(beta(j,k))+log_pr_gi(i,j,k)
       enddo
     enddo
   enddo
   
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns log-probability of data
 function log_pr_augm_data()
   double precision :: log_pr_augm_data
   integer :: i
   
   log_pr_augm_data=0.D0
   do i=1,nind
     iprob(i)=log_pr_gi(i,ianc(i,1),ianc(i,2))
     log_pr_augm_data=log_pr_augm_data+iprob(i)
   enddo
   
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws ancestry for seed and pollen gametes
 ! two hacks are used to cope with potential under-flow: const_pop and const_loc
 subroutine rnd_ancestry_i(i,tjsr)
   integer :: i,j,k
   integer*8, intent(inout) :: tjsr
   double precision :: r,sum_pr,tot_pr,pr(npop,npop),const_loc,const_pop
   
   const_pop=dble(npop*npop)
   
   ianc(i,:)=ipop(i)
   const_loc=0.D0 ! the constant equal to the max log prob. pr_gi(i,j,k)
   do j=1,npop
     do k=1,npop
       pr(j,k)=log_pr_gi(i,j,k)
       if (j==1) then
         const_loc=pr(j,k)
       else
         if (pr(j,k)>const_loc) const_loc=pr(j,k)
       endif
     enddo
   enddo
   
   tot_pr=0.D0
   do j=1,npop
     do k=1,npop
       pr(j,k)=const_pop*alpha(ipop(i),j)*beta(j,k)*dexp(pr(j,k)-const_loc+500.D0)
       tot_pr=tot_pr+pr(j,k)
     enddo
   enddo
   
   r=prnd(tjsr)
   sum_pr=0.D0
   do j=1,npop
     do k=1,npop
       sum_pr=sum_pr+pr(j,k)/tot_pr
       if (r<sum_pr) then
         ianc(i,1)=j
         ianc(i,2)=k
         return
       endif
     enddo
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws auxiliary data (unobserved variables): allele ancestry & dropout events
 subroutine rnd_alleles_il(i,l,j,k,tjsr)
   integer :: i,l,j,k
   integer*8, intent(inout) :: tjsr
   double precision :: r,p1,p2,p3
   
   if (igen(i,l,1) /= -1 .and. igen(i,l,2) /= -1) then
     r=prnd(tjsr)
     if (igen(i,l,1)==igen(i,l,2)) then
       ! homozygote
       if (j==k) then
         ! alleles from the same locality
         p1=(1.D0-ri(i))*pklm(j,l,igen(i,l,1)) ! true non-ibd homozygote
         p2=ri(i) ! true ibd homozygote
         p3=(1.D0-ri(i))*dl(l)*(1.D0-pklm(j,l,igen(i,l,1))) ! true heterozygote
         if (r<p1/(p1+p2+p3)) then
           ! true non-ibd homozygote
           igenanc(i,l,1)=j
           igenanc(i,l,2)=j
           iinb(i,2)=iinb(i,2)+1
         else if (r<(p1+p2)/(p1+p2+p3)) then
         ! true ibd homozygote
           igenanc(i,l,1)=j
           iinb(i,1)=iinb(i,1)+1
         else
           ! false homozygote
           igenanc(i,l,1)=j ! second allele's ancestry not specified because it's identity is not assigned
           idout(i,l)=1
           iinb(i,2)=iinb(i,2)+1
         endif
       else
         ! alleles from different localities
         iinb(i,2)=iinb(i,2)+1
         p1=pklm(j,l,igen(i,l,1))*pklm(k,l,igen(i,l,1)) ! true homozygote
         p2=0.5D0*dl(l)*pklm(j,l,igen(i,l,1))*(1.D0-pklm(k,l,igen(i,l,1))) ! true heterozygote 1
         p3=0.5D0*dl(l)*pklm(k,l,igen(i,l,1))*(1.D0-pklm(j,l,igen(i,l,1))) ! true heterozygote 2
         if (r<p1/(p1+p2+p3)) then
           ! true non-ibd homozygote
           igenanc(i,l,1)=j
           igenanc(i,l,2)=k
         else if (r<(p1+p2)/(p1+p2+p3)) then
           ! false homozygote, known allele with ancestry in j-th pop
           igenanc(i,l,1)=j
           idout(i,l)=1
         else
           ! false homozygote, known allele with ancestry in k-th pop
           igenanc(i,l,1)=k
           idout(i,l)=1
         endif
       endif
     else
       ! true heterozygote (no drop-out, no ibd)
       idout(i,l)=2
       iinb(i,2)=iinb(i,2)+1
       if (j==k) then
         ! alleles from the same locality
         igenanc(i,l,1)=j
         igenanc(i,l,2)=j
       else
         ! alleles from different localities
         p1=pklm(j,l,igen(i,l,1))*pklm(k,l,igen(i,l,2))
         p2=pklm(k,l,igen(i,l,1))*pklm(j,l,igen(i,l,2))
         if (r<p1/(p1+p2)) then
           igenanc(i,l,1)=j
           igenanc(i,l,2)=k
         else
           igenanc(i,l,1)=k
           igenanc(i,l,2)=j
         endif
       endif
     endif
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws auxiliary data (unobserved variables) for i-th individual
 subroutine draw_aux_i(i,tjsr)
   integer :: i,l
   integer*8, intent(inout) :: tjsr
   
   if (.not. ideal_data) call rnd_ancestry_i(i,tjsr)
   do l=1,nloc
     call rnd_alleles_il(i,l,ianc(i,1),ianc(i,2),tjsr)
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws auxiliary data (unobserved variables)
 subroutine draw_aux
   integer :: i,l,th
   integer*8 :: tjsr
   
   nak=0
   nbk=0
   if (.not. ideal_data) ianc=0
   igenanc=0
   nklm=0
   ndl=0
   idout=0
   iinb=0
   
   !$omp parallel num_threads(nthreads) default(shared) private(th,tjsr)
   !$omp do
   do i=1,nind
     th=omp_get_thread_num()+1
     tjsr=pjsr(th)
     call draw_aux_i(i,tjsr)
     pjsr(th)=tjsr
   enddo
   !$omp end do
   !$omp end parallel
   
   do i=1,nind
     nak(ipop(i),ianc(i,1))=nak(ipop(i),ianc(i,1))+1
     nbk(ianc(i,1),ianc(i,2))=nbk(ianc(i,1),ianc(i,2))+1
     do l=1,nloc
       if (igenanc(i,l,1)>0) nklm(igenanc(i,l,1),l,igen(i,l,1))=nklm(igenanc(i,l,1),l,igen(i,l,1))+1
       if (igenanc(i,l,2)>0) nklm(igenanc(i,l,2),l,igen(i,l,2))=nklm(igenanc(i,l,2),l,igen(i,l,2))+1
       if (idout(i,l)>0) ndl(l,idout(i,l))=ndl(l,idout(i,l))+1
     enddo
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws allelic drop-out rates
 subroutine draw_dl
   integer :: l
   double precision :: a1,a2,p1,p2
   
   !p1=m_d*(1.D0-y_d)/y_d
   !p2=(1.D0-m_d)*(1.D0-y_d)/y_d
   p1=0.01D0
   p2=0.01D0
   do l=1,nloc
     a1=gammadev(ndl(l,1)+p1)
     a2=gammadev(ndl(l,2)+p2)
     if (a1+a2 > 1.D-100) dl(l)=a1/(a1+a2)
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws alphas
 subroutine draw_alpha
   integer :: j,k
   double precision :: sa,ak(npop),pk(npop)
   
   do j=1,npop
     pk(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-y_alpha)/y_alpha)
     sa=0.D0
     do k=1,npop
       if (k/=j) pk(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
       ak(k)=gammadev(dble(nak(j,k))+pk(k))
       sa=sa+ak(k)
     enddo
     
     ! 'maxval' condition mitigates label-switching problem
     ! (not sure if really needed because label-switching might be
     ! a signal of intensive migration; simulations are needed
     ! to see how often the issue occurs) 
     
     !if (maxval(ak)==ak(j) .and. sa>1.D-100) alpha(j,:)=ak/sa
     
     if (sa>1.D-100) alpha(j,:)=ak/sa
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws betas directly from the posterior
 subroutine draw_beta
   integer :: j,k
   double precision :: sa,ak(npop),pk(npop)
   
   do j=1,npop
     pk(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-y_beta)/y_beta)
     sa=0.D0
     do k=1,npop
       if (k/=j) pk(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
       ak(k)=gammadev(dble(nbk(j,k))+pk(k))
       sa=sa+ak(k)
     enddo
     
     ! 'maxval' condition can mitigate label-switching problem
     ! (not sure if really needed because label-switching might be
     ! a signal of intensive migration; simulations are needed
     ! to see how often the issue occurs) 
     !if (maxval(ak)==ak(j) .and. sa>1.D-100) beta(j,:)=ak/sa
     
     if (sa>1.D-100) beta(j,:)=ak/sa
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws inbreeding for the i-th individual directly from the posterior
 subroutine draw_r(i,tjsr)
   integer :: i
   double precision :: sa,a1,pk(2),fyr
   integer*8, intent(inout) :: tjsr
   
   if (ianc(i,1)==ianc(i,2) .and. ianc(i,1)/=0) then
     fyr=((1.D0-y_r)/y_r)
     pk(1)=m_r*fyr
     pk(2)=(1.D0-m_r)*fyr
     sa=0.D0
     a1=pgammadev(dble(iinb(i,1))+pk(1),tjsr)
     sa=a1+pgammadev(dble(iinb(i,2))+pk(2),tjsr)
     if (sa>1.D-100) ri(i)=a1/sa
   else
     ri(i)=0.D0
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws inbreeding rates
 subroutine draw_ri
   integer :: k,th
   integer*8 :: tjsr
   
   !$omp parallel num_threads(nthreads) default(shared) private(th,tjsr)
   !$omp do
   do k=1,nind
     th=omp_get_thread_num()+1
     tjsr=pjsr(th)
     call draw_r(k,tjsr)
     pjsr(th)=tjsr
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws allele freqs in k-th pop
 subroutine draw_pl(l,tjsr)
   integer :: k,l,m
   double precision :: sa,am(maxnall),pm(maxnall)
   integer*8, intent(inout) :: tjsr
   
   do k=1,npop
     sa=0.D0
     do m=1,nall(l)
       pm(m)=(dble(1)/fk(k)-dble(1))*qlm(l,m)
       am(m)=pgammadev(dble(nklm(k,l,m))+pm(m),tjsr)
       sa=sa+am(m)
     enddo
     if (sa>1.D-100) pklm(k,l,1:nall(l))=am(1:nall(l))/sa
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws allele freqs in k-th pop
 subroutine draw_pk(k,tjsr)
   integer :: k,l,m
   double precision :: sa,am(maxnall),pm(maxnall)
   integer*8, intent(inout) :: tjsr
   
   if (npop==1) then
     do l=1,nloc
       sa=0.D0
       do m=1,nall(l)
         pm(m)=dble(1)
         am(m)=pgammadev(dble(nklm(k,l,m))+pm(m),tjsr)
         sa=sa+am(m)
       enddo
       if (sa>1.D-100) pklm(k,l,1:nall(l))=am(1:nall(l))/sa
     enddo
   else
     do l=1,nloc
       sa=0.D0
       do m=1,nall(l)
         pm(m)=(dble(1)/fk(k)-dble(1))*qlm(l,m)
         am(m)=pgammadev(dble(nklm(k,l,m))+pm(m),tjsr)
         sa=sa+am(m)
       enddo
       if (sa>1.D-100) pklm(k,l,1:nall(l))=am(1:nall(l))/sa
     enddo
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws allele freqs in pops
 subroutine draw_pklm
   integer :: k,th,nth
   integer*8 :: tjsr
   
   nth=min(nloc,npop,nthreads)
   
   if (nloc<npop) then
     !$omp parallel num_threads(nth) default(shared) private(th,tjsr)
     !$omp do
     do k=1,nloc
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       call draw_pl(k,tjsr)
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
   else
     !$omp parallel num_threads(nth) default(shared) private(th,tjsr)
     !$omp do
     do k=1,npop
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       call draw_pk(k,tjsr)
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws ancestral allele freqs in j-th locus
 subroutine draw_ql(j,tjsr)
   integer, intent(in) :: j
   integer :: a1,a2,k
   double precision :: xo1,xo2,difl,tk,b1o,b2o,b1n,b2n
   integer*8, intent(inout) :: tjsr
   
   if (nall(j)==1) return
   
   a1 = int(prnd(tjsr)*nall(j)) + 1
   a2 = int(prnd(tjsr)*(nall(j)-1)) + 1
   if (a2==a1) a2=a2+1
   xo1 = qlm(j,a1)
   xo2 = qlm(j,a2)
   
   qlm(j,a1) = prectdev(xo1,delta_q,tjsr)
   qlm(j,a2) = xo2 - (qlm(j,a1) - xo1)
   
   if (qlm(j,a2) <= MINQLM .or. qlm(j,a2) >= MAXQLM) then
     qlm(j,a1) = xo1
     qlm(j,a2) = xo2
     reject_q=reject_q+1.D0
     return
   endif
   
   difl=dble(0)
   do k=1,npop
     tk=(dble(1)/fk(k)-dble(1))
     b1o=tk*xo1
     b2o=tk*xo2
     b1n=tk*qlm(j,a1)
     b2n=tk*qlm(j,a2)
     
     difl=difl+(b1n-b1o)*dlog(pklm(k,j,a1))+(b2n-b2o)*dlog(pklm(k,j,a2)) &
     +DLGAMA(b1o)+DLGAMA(b2o)-DLGAMA(b1n)-DLGAMA(b2n)
   enddo
   
   if (dlog(prnd(tjsr))>=(difl)) then
     qlm(j,a1) = xo1
     qlm(j,a2) = xo2
     reject_q=reject_q+1.D0
   else
     accept_q=accept_q+1.D0
   endif 
   
 end subroutine draw_ql
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws ancestral allele freqs
 subroutine draw_qlm
   integer :: k,th
   integer*8 :: tjsr
   
   if (npop==1) return
   
   !$omp parallel num_threads(nthreads) default(shared) private(th,tjsr)
   !$omp do
   do k=1,nloc
     th=omp_get_thread_num()+1
     tjsr=pjsr(th)
     call draw_ql(k,tjsr)
     pjsr(th)=tjsr
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws genetic divergence rate for k-th population
 subroutine draw_fk(k)
   integer, intent(in) :: k
   integer :: l,m
   double precision :: xo,difl,pro,prn,thn,tho
   double precision :: so0,so1,s2,sn0,sn1
   
   if (npop==1) return
   
   pro = log_GammaD(fk(k),m_f,y_f)
   
   xo=fk(k)
   fk(k)=rectdev(xo,delta_f)
   
   if (fk(k)<MINFK .or. fk(k)>MAXFK) then
     fk(k)=xo
     reject_f=reject_f+1.D0
     return
   endif
   
   thn=dble(1)/fk(k)-dble(1)
   tho=dble(1)/xo-dble(1)
   so0=dble(nloc)*DLGAMA(tho)
   so1=dble(0)
   s2=dble(0)
   sn0=dble(nloc)*DLGAMA(thn)
   sn1=dble(0)
   difl=0.D0
   do l=1,nloc
     do m=1,nall(l)
       so1=so1+DLGAMA(tho*qlm(l,m))
       s2=s2+qlm(l,m)*dlog(pklm(k,l,m))
       sn1=sn1+DLGAMA(thn*qlm(l,m))
     enddo
   enddo
   
   prn = log_GammaD(fk(k),m_f,y_f)
   
   difl=sn0-sn1+thn*s2-so0+so1-tho*s2
   
   if (dlog(rnd())>=(difl + prn - pro)) then
     fk(k) = xo
     reject_f=reject_f+1.D0
   else
     accept_f=accept_f+1.D0
   endif
  
 end subroutine draw_fk
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws hyper-parameters for genetic divergence rates
 subroutine draw_my_f
   double precision :: lo,ln,xo!,xn
   integer :: i
   
   if (npop==1) return
   
   xo=m_f
   m_f=rectdev(xo,delta_mf)
   lo=dble(0)
   ln=dble(0)
   do i=1,npop
     lo=lo+log_GammaD(fk(i),xo,y_f)
     ln=ln+log_GammaD(fk(i),m_f,y_f)
   enddo
   
   if (dlog(rnd()).ge.(ln-lo)) then
     m_f = xo
     reject_mf=reject_mf+1.D0
   else
     accept_mf=accept_mf+1.D0
     !lo=ln
   endif
   
   ! updating the shape would likely fail to generate reasonable results based on a limited number of pops
   !
   !xo=yf
   !xn=rectdev(xo/(dble(1)-xo),rectwdth)
   !yf=xn/(dble(1)+xn)
   !ln=dble(0)
   !do i=1,nlin
   !  ln=ln+log_GammaD(fk(i),xo,yf)
   !enddo
   !
   !if (dlog(uniformdev()).ge.(ln-lo)) then
   !  yf = xo
   !endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws hyper-parameters for inbreeding rates
 subroutine draw_my_r
   double precision :: lo,ln,xo
   integer :: i
   double precision, allocatable :: bo(:),bn(:),x(:)
   
   allocate(bo(2))
   allocate(bn(2))
   allocate(x(2))
   
   xo=m_r
   bo(1)=m_r*(dble(1)-y_r)/y_r
   bo(2)=(dble(1)-m_r)*(dble(1)-y_r)/y_r
   lo=dble(0)
   
   m_r=rectdev(xo,delta_mr)
   bn(1)=m_r*(dble(1)-y_r)/y_r
   bn(2)=(dble(1)-m_r)*(dble(1)-y_r)/y_r
   ln=dble(0)
   do i=1,nind
     if (ianc(i,1)/=ianc(i,2)) cycle
     x(1)=ri(i)
     x(2)=dble(1)-ri(i)
     lo = lo + log_DirichletD(x,bo,2)
     ln = ln + log_DirichletD(x,bn,2)
   enddo
   
   if (dlog(rnd()).ge.(ln - lo)) then
     m_r = xo
     reject_mr=reject_mr+1.D0
   else
     accept_mr=accept_mr+1.D0
     lo=ln
   endif
   
   xo=y_r
   y_r=rectdev(xo,delta_yr)
   if (y_r<MINYR .or. y_r>MAXYR) then
     y_r=xo
     reject_yr=reject_yr+1.D0
     return
   endif
   bn(1)=m_r*(dble(1)-y_r)/y_r
   bn(2)=(dble(1)-m_r)*(dble(1)-y_r)/y_r
   ln=dble(0)
   do i=1,nind
     if (ianc(i,1)/=ianc(i,2)) cycle
     x(1)=ri(i)
     x(2)=dble(1)-ri(i)
     ln = ln + log_DirichletD(x,bn,2)
   enddo
   
   if (dlog(rnd()).ge.(ln - lo)) then
     y_r = xo
     reject_yr=reject_yr+1.D0
   else
     accept_yr=accept_yr+1.D0
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws isolation parameter for alphas
 subroutine draw_z_alpha
   double precision :: lo,ln,xo
   integer :: j,k
   double precision, allocatable :: b(:)
   
   if (npop==1) return
   
   allocate(b(npop))
   
   xo=z_alpha
   z_alpha=rectdev(xo,delta_za)
   if (z_alpha<MINZA .or. z_alpha>MAXZA) then
     z_alpha=xo
     reject_za=reject_za+1.D0
     return
   endif
   lo=0.D0
   ln=0.D0
   do j=1,npop
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_alpha)*(1.D0-xo)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
     enddo
     b(j)=(l_alpha+(1.D0-l_alpha)*xo)*((1.D0-y_alpha)/y_alpha)
     lo=lo+log_DirichletD(alpha(j,:),b,npop)
     
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
     enddo
     b(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-y_alpha)/y_alpha)
     ln=ln+log_DirichletD(alpha(j,:),b,npop)
   enddo
   
   if (dlog(rnd()).ge.(ln - lo)) then
     z_alpha = xo
     reject_za=reject_za+1.D0
   else
     accept_za=accept_za+1.D0
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws isolation parameter for betas
 subroutine draw_z_beta
   double precision :: lo,ln,xo
   integer :: j,k
   double precision, allocatable :: b(:)
   
   if (npop==1) return
   
   allocate(b(npop))
   
   xo=z_beta
   z_beta=rectdev(xo,delta_zb)
   if (z_beta<MINZB .or. z_beta>MAXZB) then
     z_beta=xo
     reject_zb=reject_zb+1.D0
     return
   endif
   lo=0.D0
   ln=0.D0
   do j=1,npop
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_beta)*(1.D0-xo)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
     enddo
     b(j)=(l_beta+(1.D0-l_beta)*xo)*((1.D0-y_beta)/y_beta)
     lo=lo+log_DirichletD(beta(j,:),b,npop)
     
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
     enddo
     b(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-y_beta)/y_beta)
     ln=ln+log_DirichletD(beta(j,:),b,npop)
   enddo
   
   if (dlog(rnd()).ge.(ln - lo)) then
     z_beta = xo
     reject_zb=reject_zb+1.D0
   else
     accept_zb=accept_zb+1.D0
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws dispersion parameter for alphas
 subroutine draw_y_alpha
   double precision :: lo,ln,xo
   integer :: j,k
   double precision, allocatable :: b(:)
   
   if (npop==1) return
   
   allocate(b(npop))
   
   xo=y_alpha
   y_alpha=rectdev(xo,delta_ya)
   if (y_alpha<MINYA .or. y_alpha>MAXYA) then
     y_alpha=xo
     reject_ya=reject_ya+1.D0
     return
   endif
   lo=0.D0
   ln=0.D0
   do j=1,npop
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-xo)/xo)
     enddo
     b(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-xo)/xo)
     lo=lo+log_DirichletD(alpha(j,:),b,npop)
     
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
     enddo
     b(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-y_alpha)/y_alpha)
     ln=ln+log_DirichletD(alpha(j,:),b,npop)
   enddo
   
   if (dlog(rnd()).ge.(ln - lo)) then
     y_alpha = xo
     reject_ya=reject_ya+1.D0
   else
     accept_ya=accept_ya+1.D0
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws dispersion parameter for betas
 subroutine draw_y_beta
   double precision :: lo,ln,xo
   integer :: j,k
   double precision, allocatable :: b(:)
   
   if (npop==1) return
   
   allocate(b(npop))
   
   xo=y_beta
   y_beta=rectdev(xo,delta_yb)
   if (y_beta<MINYB .or. y_beta>MAXYB) then
     y_beta=xo
     reject_yb=reject_yb+1.D0
     return
   endif
   lo=0.D0
   ln=0.D0
   do j=1,npop
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-xo)/xo)
     enddo
     b(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-xo)/xo)
     lo=lo+log_DirichletD(beta(j,:),b,npop)
     
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
     enddo
     b(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-y_beta)/y_beta)
     ln=ln+log_DirichletD(beta(j,:),b,npop)
   enddo
   
   if (dlog(rnd()).ge.(ln - lo)) then
     y_beta = xo
     reject_yb=reject_yb+1.D0
   else
     accept_yb=accept_yb+1.D0
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws spatial parameter for alphas
 subroutine draw_b_alpha
   double precision :: lo,ln,xo,po,pn
   integer :: j,k
   double precision, allocatable :: b(:)
   
   if (npop==1 .or. incl_b_alpha==0) return
   
   allocate(b(npop))
   
   xo=b_alpha
   b_alpha=xo+normaldev(0.D0,sigma_ba)
   if (b_alpha<MINBA .or. b_alpha>MAXBA) then
     b_alpha=xo
     reject_ba=reject_ba+1.D0
     return
   endif
   lo=0.D0
   ln=0.D0
   po=log_NormD(xo,0.D0,v_b)
   pn=log_NormD(b_alpha,0.D0,v_b)
   
   do j=1,npop
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
     enddo
     b(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-y_alpha)/y_alpha)
     lo=lo+log_DirichletD(alpha(j,:),b,npop)
   enddo
   
   call compute_r_alpha
   do j=1,npop
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
     enddo
     b(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-y_alpha)/y_alpha)
     ln=ln+log_DirichletD(alpha(j,:),b,npop)
   enddo
   
   if (dlog(rnd()).ge.(ln - lo + pn - po)) then
     b_alpha = xo
     call compute_r_alpha
     reject_ba=reject_ba+1.D0
   else
     accept_ba=accept_ba+1.D0
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws random spatial parameter for betas
 subroutine draw_b_beta
   double precision :: lo,ln,xo,po,pn
   integer :: j,k
   double precision, allocatable :: b(:)
   
   if (npop==1 .or. incl_b_beta==0) return
   
   allocate(b(npop))
   
   xo=b_beta
   b_beta=xo+normaldev(0.D0,sigma_bb)
   if (b_beta<MINBB .or. b_beta>MAXBB) then
     b_beta=xo
     reject_bb=reject_bb+1.D0
     return
   endif
   lo=0.D0
   ln=0.D0
   po=log_NormD(xo,0.D0,v_b)
   pn=log_NormD(b_beta,0.D0,v_b)
   
   do j=1,npop
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
     enddo
     b(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-y_beta)/y_beta)
     lo=lo+log_DirichletD(beta(j,:),b,npop)
   enddo
   
   call compute_r_beta
   do j=1,npop
     do k=1,npop
       if (k==j) cycle
       b(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
     enddo
     b(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-y_beta)/y_beta)
     ln=ln+log_DirichletD(beta(j,:),b,npop)
   enddo
   
   if (dlog(rnd()).ge.(ln - lo + pn - po)) then
     b_beta = xo
     call compute_r_beta
     reject_bb=reject_bb+1.D0
   else
     accept_bb=accept_bb+1.D0
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes spatial model used as expectation for alphas
 subroutine compute_r_alpha
   integer :: k,l
   double precision :: sa
   
   if (npop==1) return
   
   do k=1,npop
     r_alpha(k,k)=0.D0
     sa=r_alpha(k,k)
     do l=1,npop
       if (l==k) cycle
       !r_alpha(k,l)=dexp(-b_alpha*dlog(1.D0+pdist(k,l))) ! equivalent expression, but given as an exponential func.
       r_alpha(k,l)=1.D0/((1.D0+pdist(k,l))**b_alpha)
       sa=sa+r_alpha(k,l)
     enddo
     r_alpha(k,:)=r_alpha(k,:)/sa
   enddo
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes spatial model used as expectation for betas
 subroutine compute_r_beta
   integer :: k,l
   double precision :: sb
   
   if (npop==1) return
   
   do k=1,npop
     r_beta(k,k)=0.D0
     sb=r_beta(k,k)
     do l=1,npop
       if (l==k) cycle
       !r_beta(k,l)=dexp(-b_beta*dlog(1.D0+pdist(k,l))) ! equivalent expression, but given as an exponential func.
       r_beta(k,l)=1.D0/((1.D0+pdist(k,l))**b_beta)
       sb=sb+r_beta(k,l)
     enddo
     r_beta(k,:)=r_beta(k,:)/sb
   enddo
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine makes a jump between the spatial and non-spatial model for alpha
 subroutine update_alpha_model
   integer :: j,k
   double precision :: lo,ln,po,pn,xo,b(npop)
   
   ! lo - loglikelihood before parameter change
   ! ln - loglikelihood after parameter change
   ! xo - parameter value before change
   ! b(:) - vector of Dirichlet parameters
   
   if (incl_b_alpha==1) then
     ! try removing spatial effect from the migration model
     lo=0.D0
     ln=0.D0
     xo=b_alpha
     
     do j=1,npop
       do k=1,npop
         if (k==j) cycle
         b(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
       enddo
       b(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-y_alpha)/y_alpha)
       lo=lo+log_DirichletD(alpha(j,:),b,npop)
     enddo
     po=log_NormD(xo,0.D0,v_b)
     
     b_alpha=0.D0
     call compute_r_alpha
     do j=1,npop
       do k=1,npop
         if (k==j) cycle
         b(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
       enddo
       b(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-y_alpha)/y_alpha)
       ln=ln+log_DirichletD(alpha(j,:),b,npop)
     enddo
     pn=log_NormD(b_alpha,mq_b_alpha,vq_b_alpha)
     
     if (dlog(rnd()).ge.(ln - lo + pn - po)) then
       ! failed to remove
       b_alpha=xo
       call compute_r_alpha
     else
       ! succesfully removed
       incl_b_alpha=0
     endif
   else
     ! try adding spatial effect from the migration model
     lo=0.D0
     ln=0.D0
     xo=b_alpha
     
     do j=1,npop
       do k=1,npop
         if (k==j) cycle
         b(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
       enddo
       b(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-y_alpha)/y_alpha)
       lo=lo+log_DirichletD(alpha(j,:),b,npop)
     enddo
     po=log_NormD(xo,mq_b_alpha,vq_b_alpha)
               
     do 
       b_alpha=normaldev(mq_b_alpha,dsqrt(vq_b_alpha))
       if (b_alpha>MINBA .and. b_alpha<MAXBA) exit
     enddo
     
     call compute_r_alpha
     do j=1,npop
       do k=1,npop
         if (k==j) cycle
         b(k)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(j,k)*((1.D0-y_alpha)/y_alpha)
       enddo
       b(j)=(l_alpha+(1.D0-l_alpha)*z_alpha)*((1.D0-y_alpha)/y_alpha)
       ln=ln+log_DirichletD(alpha(j,:),b,npop)
     enddo
     pn=log_NormD(b_alpha,0.D0,v_b)
     
     if (dlog(rnd()).ge.(ln - lo + pn - po)) then
       b_alpha=xo
       call compute_r_alpha
     else
       incl_b_alpha=1
     endif
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine makes a jump between the spatial and non-spatial model for beta
 subroutine update_beta_model
   integer :: j,k
   double precision :: lo,ln,po,pn,xo,b(npop)
   
   ! lo - loglikelihood before parameter change
   ! ln - loglikelihood after parameter change
   ! xo - parameter value before change
   ! b(:) - vector of Dirichlet parameters
   
   if (incl_b_beta==1) then
     lo=0.D0
     ln=0.D0
     xo=b_beta
     
     do j=1,npop
       do k=1,npop
         if (k==j) cycle
         b(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
       enddo
       b(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-y_beta)/y_beta)
       lo=lo+log_DirichletD(beta(j,:),b,npop)
     enddo
     po=log_NormD(xo,0.D0,v_b)
     
     b_beta=0.D0
     call compute_r_beta
     do j=1,npop
       do k=1,npop
         if (k==j) cycle
         b(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
       enddo
       b(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-y_beta)/y_beta)
       ln=ln+log_DirichletD(beta(j,:),b,npop)
     enddo
     pn=log_NormD(b_beta,mq_b_beta,vq_b_beta)
     
     if (dlog(rnd()).ge.(ln - lo + pn - po)) then
       b_beta=xo
       call compute_r_beta
     else
       incl_b_beta=0
     endif
   else
     lo=0.D0
     ln=0.D0
     xo=b_beta
     
     do j=1,npop
       do k=1,npop
         if (k==j) cycle
         b(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
       enddo
       b(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-y_beta)/y_beta)
       lo=lo+log_DirichletD(beta(j,:),b,npop)
     enddo
     po=log_NormD(xo,mq_b_beta,vq_b_beta)
               
     do 
       b_beta=normaldev(mq_b_beta,dsqrt(vq_b_beta))
       if (b_beta>MINBB .and. b_beta<MAXBB) exit
     enddo
     
     call compute_r_beta
     do j=1,npop
       do k=1,npop
         if (k==j) cycle
         b(k)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(j,k)*((1.D0-y_beta)/y_beta)
       enddo
       b(j)=(l_beta+(1.D0-l_beta)*z_beta)*((1.D0-y_beta)/y_beta)
       ln=ln+log_DirichletD(beta(j,:),b,npop)
     enddo
     pn=log_NormD(b_beta,0.D0,v_b)
     
     if (dlog(rnd()).ge.(ln - lo + pn - po)) then
       b_beta=xo
       call compute_r_beta
     else
       incl_b_beta=1
     endif
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes the matrix of total migration rates for genes (not individuals!)
 subroutine compute_mjk
   integer :: j,k,l
   
   mjk=0.D0
   do j=1,npop
     do k=1,npop
       mjk(j,k)=0.5D0*alpha(j,k)*(1.D0+beta(k,k))
       do l=1,npop
         if (l==k) cycle
         mjk(j,k)=mjk(j,k)+0.5D0*alpha(j,l)*beta(l,k)
       enddo
     enddo
   enddo
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine initializes MCMC simulation
 subroutine initialize_mcmc
   integer :: k,l,n
   double precision :: sa,sb
   
   ! parameters
   allocate (alpha(npop,npop))
   allocate (beta(npop,npop))
   allocate (mjk(npop,npop))
   allocate (pklm(npop,nloc,maxnall))
   allocate (qlm(nloc,maxnall))
   allocate (fk(npop))
   allocate (dl(nloc))
   allocate (ri(nind))
   allocate (r_alpha(npop,npop))
   allocate (r_beta(npop,npop))
   
   z_alpha=1.D0
   z_beta=1.D0
   b_alpha=0.D0
   b_beta=0.D0
   incl_b_alpha=0
   incl_b_beta=0
   if (spatial_zygotic) incl_b_alpha=1
   if (spatial_gametic) incl_b_beta=1
   
   do k=1,npop
     r_alpha(k,k)=0.D0
     r_beta(k,k)=0.D0
     sa=0.D0
     sb=0.D0
     do l=1,npop
       if (l==k) cycle
       r_alpha(k,l)=dexp(-b_alpha*dlog(1.D0+pdist(k,l)))
       r_beta(k,l)=dexp(-b_beta*dlog(1.D0+pdist(k,l)))
       sa=sa+r_alpha(k,l)
       sb=sb+r_beta(k,l)
     enddo
     if (sa>0.D0) r_alpha(k,:)=r_alpha(k,:)/sa
     if (sb>0.D0) r_beta(k,:)=r_beta(k,:)/sb
   enddo
   
   if (zygotic_migration) z_alpha=0.5D0
   if (gametic_migration) z_beta=0.5D0
   do k=1,npop
     if (zygotic_migration) then
       alpha(k,k)=(l_alpha+(1.D0-l_alpha)*z_alpha)
     else
       alpha(k,k)=1.D0
     endif
     
     if (gametic_migration) then
       beta(k,k)=(l_beta+(1.D0-l_beta)*z_beta)
     else
       beta(k,k)=1.D0
     endif
     do l=1,npop
       if (l==k) cycle
       if (zygotic_migration) then
         alpha(k,l)=(1.D0-l_alpha)*(1.D0-z_alpha)*r_alpha(k,l)
       else
         alpha(k,l)=0.D0
       endif
       
       if (gametic_migration) then
         beta(k,l)=(1.D0-l_beta)*(1.D0-z_beta)*r_beta(k,l)
       else
         beta(k,l)=0.D0
       endif
     enddo
   enddo
   
   y_alpha=0.1D0
   y_beta=0.1D0
   
   m_f=0.2D0
   y_f=1.D0
   fk=m_f
   
   dl=0.D0
   if (dropout) dl=0.05D0
   
   m_r=0.D0
   if (inbreeding) m_r=0.05D0
   y_r=0.1D0
   ri=m_r
   do l=1,nloc
     qlm(l,:)=rlm(l,:) ! initial allele freqs metapop = observed (post-migration) mean allele freqs
     do k=1,npop
       pklm(k,l,:)=rklm(k,l,:) ! initial allele freqs within pops = observed (post-migration) allele freqs
     enddo
   enddo
   
   ! auxiliary variables
   allocate (nklm(npop,nloc,maxnall))
   allocate (nak(npop,npop))
   allocate (nbk(npop,npop))
   if (.not. ideal_data) allocate (ianc(nind,2))
   allocate (n_ianc(nind,npop,npop))
   allocate (igenanc(nind,nloc,2))
   allocate (idout(nind,nloc))
   allocate (iinb(nind,2))
   allocate (ndl(nloc,2))
   n_ianc=0
   
   ! proposal distributions
   accept_q=0.D0; reject_q=0.D0; delta_q=0.1D0
   accept_f=0.D0; reject_f=0.D0; delta_f=0.1D0
   accept_mf=0.D0; reject_mf=0.D0; delta_mf=0.1D0
   accept_md=0.D0; reject_md=0.D0; delta_md=0.1D0
   accept_mr=0.D0; reject_mr=0.D0; delta_mr=0.1D0
   accept_yr=0.D0; reject_yr=0.D0; delta_yr=0.1D0
   accept_ya=0.D0; reject_ya=0.D0; delta_ya=0.1D0
   accept_yb=0.D0; reject_yb=0.D0; delta_yb=0.1D0
   accept_za=0.D0; reject_za=0.D0; delta_za=0.1D0
   accept_zb=0.D0; reject_zb=0.D0; delta_zb=0.1D0
   accept_ba=0.D0; reject_ba=0.D0; sigma_ba=0.2D0
   accept_bb=0.D0; reject_bb=0.D0; sigma_bb=0.2D0
   
   ! means and variances across mcmc
   mlogPrAugData=0.D0; vlogPrAugData=0.D0
   allocate(mpklm(npop,nloc,maxnall));allocate(malpha(npop,npop));allocate(mbeta(npop,npop));allocate(mdl(nloc));allocate(mri(nind))
   mpklm=0.D0; malpha=0.D0; mbeta=0.D0; mdl=0.D0
   allocate(vpklm(npop,nloc,maxnall));allocate(valpha(npop,npop));allocate(vbeta(npop,npop));allocate(vdl(nloc));allocate(vri(nind))
   vpklm=0.D0; valpha=0.D0; vbeta=0.D0; vdl=0.D0
   allocate(mmjk(npop,npop)); allocate(vmjk(npop,npop))
   mmjk=0.D0; vmjk=0.D0
   my_alpha=0.D0; my_beta=0.D0; mz_alpha=0.D0; mz_beta=0.D0; mm_f=0.D0; my_f=0.D0; mm_r=0.D0; my_r=0.D0 
   vy_alpha=0.D0; vy_beta=0.D0; vz_alpha=0.D0; vz_beta=0.D0; vm_f=0.D0; vy_f=0.D0; vm_r=0.D0; vy_r=0.D0
   allocate(mqlm(nloc,maxnall)); allocate(mfk(npop))
   mqlm=0.D0; mfk=0.D0
   allocate(vqlm(nloc,maxnall)); allocate(vfk(npop))
   vqlm=0.D0; vfk=0.D0
   mb_alpha=0.D0; mb_beta=0.D0
   vb_alpha=0.D0; vb_beta=0.D0
   mPrSpatZyg=0.D0; mPrSpatGam=0.D0
   
   ! thinned list of parameter updates acros mcmc (for HDP intervals)
   n=ns/ke
   allocate(lalpha(npop,npop,n)); allocate(lbeta(npop,npop,n)); allocate(lmjk(npop,npop,n))
   allocate(lfk(npop,n)); allocate(ldl(nloc,n)); allocate(lri(nind,n))
   allocate(lm_r(n)); allocate(ly_r(n)); allocate(lm_f(n))
   allocate(ly_alpha(n)); allocate(lz_alpha(n)); allocate(lb_alpha(n))
   allocate(ly_beta(n)); allocate(lz_beta(n)); allocate(lb_beta(n))
   allocate(iprob(nind)); allocate(liprob(nind,n))
   call test_pr_gil
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine runs a single MCMC cycle
 subroutine mcmc_cycle
   integer :: i
   
   call draw_aux
   
   if (zygotic_migration) call draw_alpha
   if (gametic_migration) call draw_beta
   if (inbreeding) call draw_ri
   if (dropout) call draw_dl
   call draw_pklm
   call draw_qlm
   do i=1,npop
     call draw_fk(i)
   enddo
   call draw_my_f
   if (inbreeding) call draw_my_r
   if (zygotic_migration) call draw_z_alpha
   if (gametic_migration) call draw_z_beta
   if (zygotic_migration) call draw_y_alpha
   if (gametic_migration) call draw_y_beta
   if (zygotic_migration .and. spatial_zygotic) call draw_b_alpha
   if (gametic_migration .and. spatial_gametic) call draw_b_beta
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine runs MCMC simulation
 subroutine run_mcmc
   integer :: i, nt_init, out_file, nt_opt, nt_max, nt_min, no
   character(len=255) :: info,extension
   real*8 :: elapsed_time, tim0, tim1, tim_opt
   
   elapsed_time = omp_get_wtime()
   info='burn-in'
   call initialize_mcmc
   nu=0 ! initialize update counter for burnin stage
   no=0
   
   if (optimize_threads) then   
     nt_max=omp_get_max_threads()
     nt_min=1
     nt_init = nt_max
     nt_opt=nt_max
     no=1
     
     ! run a sampler under max number of threads
     nthreads=nt_max
     tim0=omp_get_wtime()
     call mcmc_cycle
     tim1=omp_get_wtime()
     if (mod(no,ke)==0) then
       call process_mcmc_output(info,no,nb,out_file)
     endif
     if (mod(no,1000)==0) call adjust_proposals
     
     tim_opt=tim1-tim0
     
     ! run a series of cycles under reduced number of threads
     do i=nt_init-1,nt_min,-1
       no=no+1
       nthreads=nthreads-1
       tim0=omp_get_wtime()
       call mcmc_cycle
       tim1=omp_get_wtime()
       
       if (mod(i,ke)==0) then
         !call process_mcmc_output(info,no,nb,out_file)
         
         ! sampling distrib. for RJMCMC is approximated by 70% of late burn-in samples
         if (i>nb/3) call update_mvq_b
       endif
       if (mod(no,1000)==0) call adjust_proposals
       
       
       if (tim1-tim0 < tim_opt) then
         tim_opt=tim1-tim0
         nt_opt=nthreads
       else
         nthreads = nt_opt
         exit
       endif
     enddo
   endif
   
   do i=no+1,nb
     call mcmc_cycle
     if (mod(i,ke)==0) then
       call process_mcmc_output(info,i,nb,out_file)
       
       ! sampling distrib. for RJMCMC is approximated by 70% of late burn-in samples
       if (i>nb/3) call update_mvq_b
     endif
     if (mod(i,1000)==0) call adjust_proposals
   enddo
   
   info='sampling'
   out_file=678
   extension='mcmc'
   output_file=output_filename(extension)
   open (unit=out_file,file=trim(output_file),action="write",status="replace")
   call process_mcmc_output(info,0,nb,out_file)
   nu=0 ! initialize update counter for sampling stage
   nu_b_alpha=0 ! initialize update counter for sampling stage (b_alpha only)
   nu_b_beta=0  ! initialize update counter for sampling stage (b_beta only)
   
   do i=1,ns
     call mcmc_cycle
     if (spatial_zygotic .and. rjmcmc) call update_alpha_model
     if (spatial_gametic .and. rjmcmc) call update_beta_model
     if (mod(i,ke)==0) then
       call update_means_variances
       call process_mcmc_output(info,i,ns,out_file)
     endif
   enddo
   close (out_file)
   elapsed_time = omp_get_wtime()-elapsed_time
   
   call write_summary_results(elapsed_time)
   call write_supplementary_results
   call write_gnuplot_script
   write (*,"(a,'[rspmi] Analysis completed without errors. Output written to (*.sum,*.mcmc,*.anc) files.                    ')", &
   advance="yes") creturn
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes on-line means and variances of sampling distributions used in RJMCMC for spatial parameters
 subroutine update_mvq_b
   double precision :: u,vunit
   
   nu=nu+1
   
   u=b_alpha 
   if (nu>1) then
     vunit=(u-mq_b_alpha-(u-mq_b_alpha)/dble(nu))*(u-mq_b_alpha)
     vq_b_alpha=(vq_b_alpha*dble(nu-2)+vunit)/dble(nu-1)
   endif
   mq_b_alpha=mq_b_alpha+(u-mq_b_alpha)/dble(nu)
   
   u=b_beta 
   if (nu>1) then
     vunit=(u-mq_b_beta-(u-mq_b_beta)/dble(nu))*(u-mq_b_beta)
     vq_b_beta=(vq_b_beta*dble(nu-2)+vunit)/dble(nu-1)
   endif
   mq_b_beta=mq_b_beta+(u-mq_b_beta)/dble(nu)
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes on-line means and variances of parameters
 subroutine update_means_variances
   double precision :: u,vunit
   integer :: j,k,l
   
   nu=nu+1
   
   ! sequence of inferred individual ancestry
   do j=1,nind
     n_ianc(j,ianc(j,1),ianc(j,2))=n_ianc(j,ianc(j,1),ianc(j,2))+1
   enddo
   
   ! Pr(Spatial_Models)
   nu_b_alpha=nu_b_alpha+incl_b_alpha
   nu_b_beta=nu_b_beta+incl_b_beta
   mPrSpatZyg=dble(nu_b_alpha)/dble(nu)
   mPrSpatGam=dble(nu_b_beta)/dble(nu)
   
   ! logPr(Augmented_Data)
   u=log_pr_augm_data()
   if (nu>1) then
     vunit=(u-mlogPrAugData-(u-mlogPrAugData)/dble(nu))*(u-mlogPrAugData)
     vlogPrAugData=(vlogPrAugData*dble(nu-2)+vunit)/dble(nu-1)
   endif
   mlogPrAugData=mlogPrAugData+(u-mlogPrAugData)/dble(nu)
   
   ! overdispersion in alphas
   u=y_alpha 
   if (nu>1) then
     vunit=(u-my_alpha-(u-my_alpha)/dble(nu))*(u-my_alpha)
     vy_alpha=(vy_alpha*dble(nu-2)+vunit)/dble(nu-1)
   endif
   my_alpha=my_alpha+(u-my_alpha)/dble(nu)
   
   ! overdispersion in betas
   u=y_beta
   if (nu>1) then
     vunit=(u-my_beta-(u-my_beta)/dble(nu))*(u-my_beta)
     vy_beta=(vy_beta*dble(nu-2)+vunit)/dble(nu-1)
   endif
   my_beta=my_beta+(u-my_beta)/dble(nu)
   
   ! isolation in alphas
   u=z_alpha 
   if (nu>1) then
     vunit=(u-mz_alpha-(u-mz_alpha)/dble(nu))*(u-mz_alpha)
     vz_alpha=(vz_alpha*dble(nu-2)+vunit)/dble(nu-1)
   endif
   mz_alpha=mz_alpha+(u-mz_alpha)/dble(nu)
   
   ! isolation in betas
   u=z_beta
   if (nu>1) then
     vunit=(u-mz_beta-(u-mz_beta)/dble(nu))*(u-mz_beta)
     vz_beta=(vz_beta*dble(nu-2)+vunit)/dble(nu-1)
   endif
   mz_beta=mz_beta+(u-mz_beta)/dble(nu)
   
   ! spatial component in alphas
   if (incl_b_alpha==1) then
     u=b_alpha 
     if (nu_b_alpha>1) then
       vunit=(u-mb_alpha-(u-mb_alpha)/dble(nu_b_alpha))*(u-mb_alpha)
       vb_alpha=(vb_alpha*dble(nu_b_alpha-2)+vunit)/dble(nu_b_alpha-1)
     endif
     mb_alpha=mb_alpha+(u-mb_alpha)/dble(nu_b_alpha)
   endif
   
   ! spatial component in betas
   if (incl_b_beta==1) then
     u=b_beta
     if (nu_b_beta>1) then
       vunit=(u-mb_beta-(u-mb_beta)/dble(nu_b_beta))*(u-mb_beta)
       vb_beta=(vb_beta*dble(nu_b_beta-2)+vunit)/dble(nu_b_beta-1)
     endif
     mb_beta=mb_beta+(u-mb_beta)/dble(nu_b_beta)
   endif
   
   ! mean inbreeding
   u=m_r
   if (nu>1) then
     vunit=(u-mm_r-(u-mm_r)/dble(nu))*(u-mm_r)
     vm_r=(vm_r*dble(nu-2)+vunit)/dble(nu-1)
   endif
   mm_r=mm_r+(u-mm_r)/dble(nu)
   
   ! overdispersion in inbreeding
   u=y_r
   if (nu>1) then
     vunit=(u-my_r-(u-my_r)/dble(nu))*(u-my_r)
     vy_r=(vy_r*dble(nu-2)+vunit)/dble(nu-1)
   endif
   my_r=my_r+(u-my_r)/dble(nu)
   
   ! mean fst
   u=m_f
   if (nu>1) then
     vunit=(u-mm_f-(u-mm_f)/dble(nu))*(u-mm_f)
     vm_f=(vm_f*dble(nu-2)+vunit)/dble(nu-1)
   endif
   mm_f=mm_f+(u-mm_f)/dble(nu)
   
   do j=1,nind
     ! individual inbreeding
     u=ri(j)
     if (nu>1) then
       vunit=(u-mri(j)-(u-mri(j))/dble(nu))*(u-mri(j))
       vri(j)=(vri(j)*dble(nu-2)+vunit)/dble(nu-1)
     endif
     mri(j)=mri(j)+(u-mri(j))/dble(nu)
   enddo
   
   ! compute gene migration matrix based on alpha/beta
   call compute_mjk
   
   do j=1,npop
     ! population fst
     u=fk(j)
     if (nu>1) then
       vunit=(u-mfk(j)-(u-mfk(j))/dble(nu))*(u-mfk(j))
       vfk(j)=(vfk(j)*dble(nu-2)+vunit)/dble(nu-1)
     endif
     mfk(j)=mfk(j)+(u-mfk(j))/dble(nu)
     
     ! population allele freqs
     do k=1,nloc
       do l=1,nall(k)
         u=pklm(j,k,l)
         if (nu>1) then
           vunit=(u-mpklm(j,k,l)-(u-mpklm(j,k,l))/dble(nu))*(u-mpklm(j,k,l))
           vpklm(j,k,l)=(vpklm(j,k,l)*dble(nu-2)+vunit)/dble(nu-1)
         endif
         mpklm(j,k,l)=mpklm(j,k,l)+(u-mpklm(j,k,l))/dble(nu)
       enddo 
     enddo
     
     do k=1,npop
       ! alphas
       u=alpha(j,k)
       if (nu>1) then
         vunit=(u-malpha(j,k)-(u-malpha(j,k))/dble(nu))*(u-malpha(j,k))
         valpha(j,k)=(valpha(j,k)*dble(nu-2)+vunit)/dble(nu-1)
       endif
       malpha(j,k)=malpha(j,k)+(u-malpha(j,k))/dble(nu)
       
       ! betas
       u=beta(j,k)
       if (nu>1) then
         vunit=(u-mbeta(j,k)-(u-mbeta(j,k))/dble(nu))*(u-mbeta(j,k))
         vbeta(j,k)=(vbeta(j,k)*dble(nu-2)+vunit)/dble(nu-1)
       endif
       mbeta(j,k)=mbeta(j,k)+(u-mbeta(j,k))/dble(nu)
       
       ! gene migration
       u=mjk(j,k)
       if (nu>1) then
         vunit=(u-mmjk(j,k)-(u-mmjk(j,k))/dble(nu))*(u-mmjk(j,k))
         vmjk(j,k)=(vmjk(j,k)*dble(nu-2)+vunit)/dble(nu-1)
       endif
       mmjk(j,k)=mmjk(j,k)+(u-mmjk(j,k))/dble(nu)
     enddo
   enddo
   
   do j=1,nloc
     ! locus drop-out
     u=dl(j)
     if (nu>1) then
       vunit=(u-mdl(j)-(u-mdl(j))/dble(nu))*(u-mdl(j))
       vdl(j)=(vdl(j)*dble(nu-2)+vunit)/dble(nu-1)
     endif
     mdl(j)=mdl(j)+(u-mdl(j))/dble(nu)
     
     ! ancestral allele frequencies
     do k=1,nall(j)
       u=qlm(j,k)
       if (nu>1) then
         vunit=(u-mqlm(j,k)-(u-mqlm(j,k))/dble(nu))*(u-mqlm(j,k))
         vqlm(j,k)=(vqlm(j,k)*dble(nu-2)+vunit)/dble(nu-1)
       endif
       mqlm(j,k)=mqlm(j,k)+(u-mqlm(j,k))/dble(nu)
     enddo
   enddo
   
   lalpha(:,:,nu)=alpha(:,:)
   lbeta(:,:,nu)=beta(:,:)
   lmjk(:,:,nu)=mjk(:,:)
   lri(:,nu)=ri(:)
   lfk(:,nu)=fk(:)
   ldl(:,nu)=dl(:)
   ly_alpha(nu)=y_alpha
   ly_beta(nu)=y_beta
   lz_alpha(nu)=z_alpha
   lz_beta(nu)=z_beta
   lm_r(nu)=m_r
   ly_r(nu)=y_r
   lm_f(nu)=m_f
   if (incl_b_alpha==1) lb_alpha(nu_b_alpha)=b_alpha
   if (incl_b_beta==1) lb_beta(nu_b_beta)=b_beta
   liprob(:,nu)=dexp(iprob)
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine processes the output
 subroutine process_mcmc_output(info,i,maxi,out_file)
   integer :: i,maxi,j,k,out_file
   character(len=255) :: info
   double precision :: logPrAData
   
   logPrAData=log_pr_augm_data()
   
   ! updates screen
   if (trim(info)=='burn-in') then
     if (spatial_zygotic .or. spatial_gametic) then
       write (*,"(4a,i3,'% ',e14.6,4(' ',f6.4),2(' ',e12.4))", advance="no") creturn,'[rspmi] ',trim(info),':',int(i*100/maxi), &
       logPrAData,z_alpha,z_beta,m_f,m_r,b_alpha,b_beta
     else
       write (*,"(4a,i3,'% ',e14.6,4(' ',f6.4))", advance="no") creturn,'[rspmi] ',trim(info),':',int(i*100/maxi), &
       logPrAData,z_alpha,z_beta,m_f,m_r
     endif
   else
     if (spatial_zygotic .or. spatial_gametic) then
       write (*,"(4a,i3,'% ',e14.6,4(' ',f6.4),2(' ',e12.4))", advance="no") creturn,'[rspmi] ',trim(info),':',int(i*100/maxi), &
       mlogPrAugData,mz_alpha,mz_beta,mm_f,mm_r,mb_alpha,mb_beta
     else
       write (*,"(4a,i3,'% ',e14.6,4(' ',f6.4))", advance="no") creturn,'[rspmi] ',trim(info),':',int(i*100/maxi), &
       mlogPrAugData,mz_alpha,mz_beta,mm_f,mm_r
     endif
   endif
   
   flush(6)
   
   ! updates output file
   if (trim(info)=='sampling') then
     if (i==0) then
       write (out_file,"(*(a))") 'stage',tab,'cycle',tab,'LogPrAugmData',(tab,trim(pname(j)),j=1,size(pname))
     else
       write (out_file,"(2a,i0,a,e16.8)",advance='no') trim(info),tab,i,tab,logPrAData
       if (zygotic_migration) write (out_file,"(*(a,f6.4))",advance='no') ((tab,alpha(j,k),k=1,npop),j=1,npop),tab,y_alpha,tab, &
        z_alpha
       if (zygotic_migration .and. spatial_zygotic) write (out_file,"(*(a,e16.8))",advance='no') tab,b_alpha
       if (gametic_migration) write (out_file,"(*(a,f6.4))",advance='no') ((tab,beta(j,k),k=1,npop),j=1,npop),tab,y_beta,tab,z_beta
       if (gametic_migration .and. spatial_gametic) write (out_file,"(*(a,e16.8))",advance='no') tab,b_beta
       if (dropout) write (out_file,"(*(a,f6.4))",advance='no') (tab,dl(k),k=1,nloc)!,tab,m_d,tab,y_d
       if (inbreeding) write (out_file,"(*(a,f6.4))",advance='no') (tab,ri(k),k=1,nind),tab,m_r,tab,y_r
       write (out_file,"(*(a,f6.4))",advance='yes') (tab,fk(k),k=1,npop),tab,m_f
       flush (out_file)
     endif
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine adjusts proposal parameters to get acceptance rate between 25 and 45%
 subroutine adjust_proposals
   double precision :: ar,fct
   
   fct=dsqrt(3.D0 - rnd())
   
   ar = accept_q / (accept_q+reject_q)
   if (ar < dble(0.25)) then 
     if (delta_q/fct<1.D0) delta_q = delta_q/fct
   else
     if (ar > dble(0.45)) then
       if (delta_q*fct<1.D0) delta_q = delta_q*fct
     endif
   endif
   accept_q = 0.D0
   reject_q = 0.D0
   
   ar = accept_f / (accept_f+reject_f)
   if (ar < dble(0.25)) then
     if (delta_f/fct<1.D0) delta_f = delta_f/fct
   else
     if (ar > dble(0.45)) then
       if (delta_f*fct<1.D0) delta_f = delta_f*fct
     endif
   endif
   accept_f = 0.D0
   reject_f = 0.D0
   
   ar = accept_mf / (accept_mf+reject_mf)
   if (ar < dble(0.25)) then
     if (delta_mf/fct<1.D0) delta_mf = delta_mf/fct
   else
     if (ar > dble(0.45)) then
       if (delta_mf*fct<1.D0) delta_mf = delta_mf*fct
     endif
   endif
   accept_mf = 0.D0
   reject_mf = 0.D0
   
   ar = accept_md / (accept_md+reject_md)
   if (ar < dble(0.25)) then
     if (delta_md/fct<1.D0) delta_md = delta_md/fct
   else
     if (ar > dble(0.45)) then
       if (delta_md*fct<1.D0) delta_md = delta_md*fct
     endif
   endif
   accept_md = 0.D0
   reject_md = 0.D0
   
   ar = accept_mr / (accept_mr+reject_mr)
   if (ar < dble(0.25)) then
     if (delta_mr/fct<1.D0) delta_mr = delta_mr/fct
   else
     if (ar > dble(0.45)) then
       if (delta_mr*fct<1.D0) delta_mr = delta_mr*fct
     endif
   endif
   accept_mr = 0.D0
   reject_mr = 0.D0
   
   ar = accept_yr / (accept_yr+reject_yr)
   if (ar < dble(0.25)) then
     if (delta_yr/fct<1.D0) delta_yr = delta_yr/fct
   else
     if (ar > dble(0.45)) then
       if (delta_yr*fct<1.D0) delta_yr = delta_yr*fct
     endif
   endif
   accept_yr = 0.D0
   reject_yr = 0.D0
   
   ar = accept_ya / (accept_ya+reject_ya)
   if (ar < dble(0.25)) then
     if (delta_ya/fct<1.D0) delta_ya = delta_ya/fct
   else
     if (ar > dble(0.45)) then
       if (delta_ya*fct<1.D0) delta_ya = delta_ya*fct
     endif
   endif
   accept_ya = 0.D0
   reject_ya = 0.D0
   
   ar = accept_yb / (accept_yb+reject_yb)
   if (ar < dble(0.25)) then
     if (delta_yb/fct<1.D0) delta_yb = delta_yb/fct
   else
     if (ar > dble(0.45)) then
       if (delta_yb*fct<1.D0) delta_yb = delta_yb*fct
     endif
   endif
   accept_yb = 0.D0
   reject_yb = 0.D0
   
   ar = accept_za / (accept_za+reject_za)
   if (ar < dble(0.25)) then
     if (delta_za/fct<1.D0) delta_za = delta_za/fct
   else
     if (ar > dble(0.45)) then
       if (delta_za*fct<1.D0) delta_za = delta_za*fct
     endif
   endif
   accept_za = 0.D0
   reject_za = 0.D0
   
   ar = accept_zb / (accept_zb+reject_zb)
   if (ar < dble(0.25)) then
     if (delta_zb/fct<1.D0) delta_zb = delta_za/fct
   else
     if (ar > dble(0.45)) then
       if (delta_zb*fct<1.D0) delta_zb = delta_zb*fct
     endif
   endif
   accept_zb = 0.D0
   reject_zb = 0.D0
   
   ar = accept_ba / (accept_ba+reject_ba)
   if (ar < dble(0.25)) then
     sigma_ba = sigma_ba/fct
   else
     if (ar > dble(0.45)) then
       sigma_ba = sigma_ba*fct
     endif
   endif
   accept_ba = 0.D0
   reject_ba = 0.D0
   
   ar = accept_bb / (accept_bb+reject_bb)
   if (ar < dble(0.25)) then
     sigma_bb = sigma_bb/fct
   else
     if (ar > dble(0.45)) then
       sigma_bb = sigma_bb*fct
     endif
   endif
   accept_bb = 0.D0
   reject_bb = 0.D0
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine writes summary output file
 subroutine write_summary_results(elapsed_time)
   integer :: i,j,k,out_file,thr,tmin,tsec,nsaved
   character(len=255) :: extension
   real*8 :: elapsed_time
   double precision :: xl,xu
   
   ! runtime in hours, minutes and seconds
   thr=int(elapsed_time/3600.D0)
   tmin=int((elapsed_time-3600.D0*thr)/60.D0)
   tsec=int(elapsed_time-3600.D0*thr-60.D0*tmin)
   
   nsaved=ns/ke
   out_file=567
   extension='sum'
   output_file=output_filename(extension)
   open (unit=out_file,file=trim(output_file),action="write",status="replace")
   
   write (out_file,"('Summary info:')")
   write (out_file,"('Options file name:',2a)") tab,trim(input_file)
   write (out_file,"('Data file name:',2a)") tab,trim(data_file)
   write (out_file,"('Number of individuals:',a,i0)") tab,nind
   write (out_file,"('Number of populations:',a,i0)") tab,npop
   write (out_file,"('Number of markers:',a,i0)") tab,nloc
   write (out_file,"('Number of alleles:',a,i0)") tab,sum(nall)
   write (out_file,"('Random seed:',a,i0)") tab,seed
   write (out_file,"('Length of burn-in stage:',a,i0)") tab,nb
   write (out_file,"('Length of sampling stage:',a,i0)") tab,ns
   write (out_file,"('Number of saved MCMC samples:',a,i0)") tab,nsaved
   write (out_file,"('Number of threads:',a,i0)") tab,nthreads
   
   if (inbreeding) then 
     write (out_file,"('Mating model:',2a)") tab,'inbreeding allowed'
   else
     write (out_file,"('Mating model:',2a)") tab,'inbreeding NOT allowed'
   endif
   if (dropout) then
     write (out_file,"('Genotyping model:',2a)") tab,'allelic drop-out allowed'
   else
     write (out_file,"('Genotyping model:',2a)") tab,'allelic drop-out NOT allowed'
   endif
   if (zygotic_migration) then
     if (spatial_zygotic) then
       write (out_file,"('Zygotic migration:',2a)") tab,'allowed AND spatially-explicit'
     else
       write (out_file,"('Zygotic migration:',2a)") tab,'allowed'
     endif
     write (out_file,"('Prior isolation term for zygotic migration:',a,f6.4)") tab,l_alpha
   else
     write (out_file,"('Zygotic migration:',2a)") tab,'NOT allowed'
   endif
   if (gametic_migration) then
     if (spatial_zygotic) then
       write (out_file,"('Gametic migration:',2a)") tab,'allowed AND spatially-explicit'
     else
       write (out_file,"('Gametic migration:',2a)") tab,'allowed'
     endif
     write (out_file,"('Prior isolation term for gametic migration:',a,f6.4)") tab,l_beta
   else
     write (out_file,"('Gametic migration:',2a)") tab,'NOT allowed'
   endif
   if ((spatial_gametic .or. spatial_zygotic) .and. rjmcmc) then
     write (out_file,"('RJMCMC for spatial component(s):',2a)") tab,'allowed'
   else
     write (out_file,"('RJMCMC for spatial component(s):',2a)") tab,'NOT allowed'
   endif
   
   write (out_file,"('Mean.LogPrAugData:',a,e16.8)") tab,mlogPrAugData
   write (out_file,"('Var.LogPrAugData:',a,e16.8)") tab,vlogPrAugData
   
   call compute_lppd(nsaved)
   write (out_file,"('WAIC:',a,e16.8)") tab,-2.D0*(lppd-pw)
   write (out_file,"('CPU time (hr:min:sec):',a,i0,':',i2.2,':',i2.2)") tab,thr,tmin,tsec
   write (out_file,"('')")
   
   write (out_file,"('Migration rates (table):')")
   if (coordinates) then
     write (out_file,"((21a))") 'recipent',tab,'source',tab,'distance',tab,'alpha',tab,'SE',tab,'HPDI_l',tab,'HPDI_u', &
     tab,'beta',tab,'SE',tab,'HPDI_l',tab,'HPDI_u'
     do i=1,npop
       do j=1,npop
         if (i/=j) then
           call hdi(lalpha(i,j,:),nsaved,0.95D0,xl,xu)
           write (out_file,"('pop_',i0,a,'pop_',i0,a,e14.6,6(a,f7.5))",advance="no") pop(i),tab,pop(j),tab,pdist(i,j),tab, &
           malpha(i,j),tab,dsqrt(valpha(i,j)),tab,xl,tab,xu,tab,mbeta(i,j),tab,dsqrt(vbeta(i,j))
           call hdi(lbeta(i,j,:),nsaved,0.95D0,xl,xu)
           write (out_file,"(2(a,f7.5))",advance="yes") tab,xl,tab,xu
         endif
       enddo
     enddo
   else
     write (out_file,"((19a))") 'recipent',tab,'source',tab,'alpha',tab,'SE',tab,'HPDI_l',tab,'HPDI_u', &
     tab,'beta',tab,'SE',tab,'HPDI_l',tab,'HPDI_u'
     do i=1,npop
       do j=1,npop
         if (i/=j) then
           call hdi(lalpha(i,j,:),nsaved,0.95D0,xl,xu)
           write (out_file,"('pop_',i0,a,'pop_',i0,6(a,f7.5))",advance="no") pop(i),tab,pop(j),tab,malpha(i,j),tab, &
           dsqrt(valpha(i,j)),tab,xl,tab,xu,tab,mbeta(i,j),tab,dsqrt(vbeta(i,j))
           call hdi(lbeta(i,j,:),nsaved,0.95D0,xl,xu)
           write (out_file,"(2(a,f7.5))",advance="yes") tab,xl,tab,xu
         endif
       enddo
     enddo
   endif
   write (out_file,"('')")
   
   write (out_file,"('Seed pool composition:')")
   write (out_file,"('recipient',*(2a,i0))") (tab,'pop_',pop(i),i=1,npop)
   do i=1,npop
     write (out_file,"('pop_',i0,*(a,f7.5))") pop(i),(tab,malpha(i,j),j=1,npop)
   enddo
   write (out_file,"('')")
   
   write (out_file,"('Pollen pool composition:')")
   write (out_file,"('recipient',*(2a,i0))") (tab,'pop_',pop(i),i=1,npop)
   do i=1,npop
     write (out_file,"('pop_',i0,*(a,f7.5))") pop(i),(tab,mbeta(i,j),j=1,npop)
   enddo
   write (out_file,"('')")
   
   write (out_file,"('Divergence rates:')")
   write (out_file,"((9a))") 'pop',tab,'fst',tab,'SE',tab,'HPDI_l',tab,'HPDI_u'
   do i=1,npop
     call hdi(lfk(i,:),nsaved,0.95D0,xl,xu)
     write (out_file,"('pop_',i0,4(a,f7.5))") pop(i),tab,mfk(i),tab,dsqrt(vfk(i)),tab,xl,tab,xu
   enddo
   write (out_file,"('')")
   
   write (out_file,"('Allele frequencies:')")
   write (out_file,"(5a,*(2a,i0))") 'marker',tab,'allele',tab,'anc',(tab,'pop_',k,k=1,npop)
   do i=1,nloc
     do j=1,nall(i)
       write (out_file,"('loc_',i0,a,i0,*(a,f7.5))") i,tab,allele(i,j),tab,mqlm(i,j),(tab,mpklm(k,i,j),k=1,npop)
     enddo
   enddo
   write (out_file,"('')")
   
   if (dropout) then
     write (out_file,"('Allelic drop-out rates:')")
     write (out_file,"(9a)") 'marker',tab,'dout',tab,'SE',tab,'HPDI_l',tab,'HPDI_u'
     do i=1,nloc
       call hdi(ldl(i,:),nsaved,0.95D0,xl,xu)
       write (out_file,"('loc_',i0,4(a,f7.5))") i,tab,mdl(i),tab,dsqrt(vdl(i)),tab,xl,tab,xu
     enddo
     write (out_file,"('')")
   endif
   
   if (inbreeding) then
     write (out_file,"('Individual inbreeding rates:')")
     write (out_file,"(11a)") 'ind',tab,'pop',tab,'fis',tab,'SE',tab,'HPDI_l',tab,'HPDI_u'
     do i=1,nind
       call hdi(lri(i,:),nsaved,0.95D0,xl,xu)
       write (out_file,"('ind_',i0,a,'pop_',i0,4(a,f7.5))") inam(i),tab,pop(ipop(i)),tab,mri(i),tab,dsqrt(vri(i)),tab,xl,tab,xu
     enddo
     write (out_file,"('')")
   endif
   
   write (out_file,"('Hyper-parameters:')")
   write (out_file,"(9a)") 'param.',tab,'estim.mean',tab,'SE',tab,'HPDI_l',tab,'HPDI_u'
   call hdi(lm_f,nsaved,0.95D0,xl,xu)
   write (out_file,"('mean_fst',4(a,f7.5))") tab,mm_f,tab,dsqrt(vm_f),tab,xl,tab,xu
   if (inbreeding) then
     call hdi(lm_r,nsaved,0.95D0,xl,xu)
     write (out_file,"('mean_fis',4(a,f7.5))") tab,mm_r,tab,dsqrt(vm_r),tab,xl,tab,xu
     call hdi(ly_r,nsaved,0.95D0,xl,xu)
     write (out_file,"('disp_fis',4(a,f7.5))") tab,my_r,tab,dsqrt(vy_r),tab,xl,tab,xu
   endif
   if (zygotic_migration) then
     call hdi(ly_alpha,nsaved,0.95D0,xl,xu)
     write (out_file,"('disp_alpha',4(a,f7.5))") tab,my_alpha,tab,dsqrt(vy_alpha),tab,xl,tab,xu
   endif
   if (gametic_migration) then
     call hdi(ly_beta,nsaved,0.95D0,xl,xu)
     write (out_file,"('disp_beta',4(a,f7.5))") tab,my_beta,tab,dsqrt(vy_beta),tab,xl,tab,xu
   endif
   if (zygotic_migration) then
     call hdi(lz_alpha,nsaved,0.95D0,xl,xu)
     write (out_file,"('isol_alpha',4(a,f7.5))") tab,mz_alpha,tab,dsqrt(vz_alpha),tab,xl,tab,xu
   endif
   if (gametic_migration) then
     call hdi(lz_beta,nsaved,0.95D0,xl,xu)
     write (out_file,"('isol_beta',4(a,f7.5))") tab,mz_beta,tab,dsqrt(vz_beta),tab,xl,tab,xu
   endif
   if (spatial_zygotic .and. rjmcmc) write (out_file,"('Pr(spat_alpha)',a,f6.4)") tab,mPrSpatZyg
   if (spatial_gametic .and. rjmcmc) write (out_file,"('Pr(spat_beta)',a,f6.4)") tab,mPrSpatGam
   if (spatial_zygotic) then
     call hdi(lb_alpha(1:nu_b_alpha),nu_b_alpha,0.95D0,xl,xu)
     write (out_file,"('spat_alpha',4(a,e16.8))") tab,mb_alpha,tab,dsqrt(vb_alpha),tab,xl,tab,xu
   endif
   if (spatial_gametic) then
     call hdi(lb_beta(1:nu_b_beta),nu_b_beta,0.95D0,xl,xu)
     write (out_file,"('spat_beta',4(a,e16.8))") tab,mb_beta,tab,dsqrt(vb_beta),tab,xl,tab,xu
   endif
   write (out_file,"('')")
   
   write (out_file,"('Acceptance rates across MCMC:')")
   write (out_file,"(5a)") 'param.',tab,'accept%',tab,'delta'
   write (out_file,"('qlm',a,f5.1,a,f6.4)") tab,100.D0*accept_q/(accept_q+reject_q),tab,delta_q
   write (out_file,"('fst',a,f5.1,a,f6.4)") tab,100.D0*accept_f/(accept_f+reject_f),tab,delta_f
   write (out_file,"('mean_fst',a,f5.1,a,f6.4)") tab,100.D0*accept_mf/(accept_mf+reject_mf),tab,delta_mf
   if (inbreeding) write (out_file,"('mean_fis',a,f5.1,a,f6.4)") tab,100.D0*accept_mr/(accept_mr+reject_mr),tab,delta_mr
   if (inbreeding) write (out_file,"('disp_fis',a,f5.1,a,f6.4)") tab,100.D0*accept_yr/(accept_yr+reject_yr),tab,delta_yr
   if (zygotic_migration) write (out_file,"('disp_alpha',a,f5.1,a,f6.4)") tab,100.D0*accept_ya/(accept_ya+reject_ya),tab,delta_ya
   if (gametic_migration) write (out_file,"('disp_beta',a,f5.1,a,f6.4)") tab,100.D0*accept_yb/(accept_yb+reject_yb),tab,delta_yb
   if (zygotic_migration) write (out_file,"('isol_alpha',a,f5.1,a,f6.4)") &
   tab,100.D0*accept_za/(accept_za+reject_za),tab,delta_za
   if (gametic_migration) write (out_file,"('isol_beta',a,f5.1,a,f6.4)") &
   tab,100.D0*accept_zb/(accept_zb+reject_zb),tab,delta_zb
   if (spatial_zygotic) write (out_file,"('spat_alpha',a,f5.1,a,e16.8)") tab,100.D0*accept_ba/(accept_ba+reject_ba),tab,sigma_ba
   if (spatial_gametic) write (out_file,"('spat_beta',a,f5.1,a,e16.8)") tab,100.D0*accept_bb/(accept_bb+reject_bb),tab,sigma_bb
   
   if (rjmcmc) then
     write (out_file,"('')")
     write (out_file,"('Parameters of proposal distributions for spatial effects:')")
     write (out_file,"(5a)") 'param.',tab,'mean',tab,'variance'
     if (spatial_zygotic) write (out_file,"('spat_alpha',2(a,e16.8))") tab,mq_b_alpha,tab,vq_b_alpha
     if (spatial_gametic) write (out_file,"('spat_beta',2(a,e16.8))") tab,mq_b_beta,tab,vq_b_beta
   endif
   
   close(out_file)
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine writes additional output files
 subroutine write_supplementary_results
   integer :: i,k,l,out_file,indq(nind,npop),indfm(nind),indhm(nind),nsaved
   character(len=255) :: extension
   double precision :: xu,xl
   
   ! posterior distribution of gene migration rates to text file
   nsaved=ns/ke
   out_file=567
   extension='mig'
   output_file=output_filename(extension)
   open (unit=out_file,file=trim(output_file),action="write",status="replace")
   
   write (out_file,"('Gene migration rates (table):')")
   write (out_file,"((11a))") 'recipent',tab,'source',tab,'migration',tab,'SE',tab,'HPDI_l',tab,'HPDI_u'
   do i=1,npop
     do k=1,npop
       call hdi(lmjk(i,k,:),nsaved,0.95D0,xl,xu)
       write (out_file,"('pop_',i0,a,'pop_',i0,6(a,f7.5))",advance="yes") pop(i),tab,pop(k),tab,mmjk(i,k),tab, &
       dsqrt(vmjk(i,k)),tab,xl,tab,xu
     enddo
   enddo
   
   write (out_file,"('')")
   write (out_file,"('Gene migration matrix:')")
   write (out_file,"('recipient',*(2a,i0))") (tab,'pop_',pop(i),i=1,npop)
   do i=1,npop
     write (out_file,"('pop_',i0,*(a,f7.5))") pop(i),(tab,mmjk(i,k),k=1,npop)
   enddo
   close(out_file)
   
   ! posterior distribution of detailed ancestry to text file
   out_file=567
   extension='anc'
   output_file=output_filename(extension)
   open (unit=out_file,file=trim(output_file),action="write",status="replace")
   
   write (out_file,"(9a)") 'Ind',tab,'Sampling loc',tab,'Seed source',tab,'Pollen source',tab,'MCMC Freq'
   
   indq=0; indfm=0; indhm=0
   do i=1,nind
     do k=1,npop
       do l=1,npop
         if (n_ianc(i,k,l)>0) write (out_file,"(i0,4(a,i0))") inam(i),tab,pop(ipop(i)),tab,pop(k),tab,pop(l),tab,n_ianc(i,k,l)
         indq(i,k)=indq(i,k)+n_ianc(i,k,l)
         indq(i,l)=indq(i,l)+n_ianc(i,k,l)
       enddo
     enddo
     indhm(i)=sum(n_ianc(i,ipop(i),:))+sum(n_ianc(i,:,ipop(i)))-2*n_ianc(i,ipop(i),ipop(i))
     indfm(i)=nu-indhm(i)-n_ianc(i,ipop(i),ipop(i))
   enddo
   close(out_file)
   
   ! posterior distribution of compact ancestry types
   out_file=678
   extension='q'
   output_file=output_filename(extension)
   open (unit=out_file,file=trim(output_file),action="write",status="replace")
   
   write (out_file,"(9a,*(2a,i0))") 'Ind',tab,'Sampling loc',tab,'NonMigrant',tab,'HalfMigrant',tab,'FullMigrant', &
   (tab,'Source',pop(k),k=1,npop)
   
   do i=1,nind
     write (out_file,"(i0,*(a,i0))") inam(i),tab,pop(ipop(i)),tab,n_ianc(i,ipop(i),ipop(i)),tab,indhm(i),tab,indfm(i), &
     (tab,indq(i,k),k=1,npop)
   enddo
   close(out_file)
   
   ! central (median) coordinates for population
   if (planexy .or. coordinates) then
     out_file=789
     extension='geo'
     output_file=output_filename(extension)
     open (unit=out_file,file=trim(output_file),action="write",status="replace")
     
     write (out_file,"(9a,*(2a,i0))") 'Pop',tab,'Lon',tab,'Lat'
     do i=1,npop
       write (out_file,"(i0,2(a,e14.6))") pop(i),tab,plon(i),tab,plat(i)
     enddo
     close(out_file)
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine implements a shift-register generator (64-bit version)
 function shift_xor(val,shift)
   integer*8 :: shift_xor
   integer*8, intent(in) :: val, shift
   
   shift_xor=ieor(val,ishft(val,shift))
 end function shift_xor
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine initializes random number generator
 subroutine init_rnd
    
    jsr=shift_xor(shift_xor(shift_xor(jsr, 13_8),-17_8),43_8)
 end subroutine init_rnd
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the uniform (0,1) pdf
 double precision function rnd()
   integer*8 :: ojsr
   ojsr=jsr
   jsr=shift_xor(shift_xor(shift_xor(jsr, 13_8),-17_8),43_8)
   rnd=dble(ojsr+jsr)*5.42101086242752217D-20+0.5D0
   
 end function rnd
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine initializes parallel random number generator
 subroutine init_prnd
   integer*8 :: i
   
   allocate(pjsr(nthreads))
   do i=1,nthreads
     pjsr(i)=int(dble(300000)*rnd())+1
     pjsr(i)=shift_xor(shift_xor(shift_xor(pjsr(i), 13_8),-17_8),43_8)
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the uniform (0,1) pdf
 ! (Parallel version)
 double precision function prnd(ujsr)
   integer*8, intent(inout) :: ujsr
   integer*8 :: ojsr
   ojsr=ujsr
   ujsr = shift_xor(shift_xor(shift_xor(ujsr, 13_8),-17_8),43_8)
   Prnd = dble(ojsr+ujsr)*5.42101086242752217D-20+0.5D0
 end function prnd
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the standard gamma pdf (for a<1)
 double precision function rgamma_alt(a)
   double precision :: a,e,x,p,b
   logical :: accept
   
   rgamma_alt=0.D0
   x=0.D0
   e=dexp(1.D0)
   b=(e+a)/e
   do
     p=rnd()*b
     accept=.true.
     if (p>1.D0) then
       x=-dlog((b-p)/a)
       if (dlog(rnd())>(a-1.D0)*dlog(x)) accept=.false.
     else
       x=dexp(dlog(p)/a)
       if (x>-dlog(rnd())) accept=.false.
     endif
     if (accept) exit
   enddo
   rgamma_alt=x
   
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the standard gamma pdf (for a>1)
 double precision function rgamma(a)
   double precision, intent(in) :: a
   double precision :: x,v,u,c,d
   
   d=a-dble(1)/dble(3)
   c=dble(1)/dsqrt(dble(9)*d)
   do
     do
       x=dsqrt(-dble(2)*dlog(rnd()))*dcos(my2pi*rnd())
       v=(dble(1)+c*x)
       if (v > dble(0)) exit
     enddo
     v=v**3
     u=rnd()
     if (u<(dble(1)-dble(0.0331)*x**4)) exit
     if (dlog(u)<(dble(0.5)*x**2+d*(dble(1)-v+dlog(v)))) exit
   enddo
   rgamma=d*v
   
 end function rgamma
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the standard gamma pdf
 double precision function gammadev(a)
   double precision :: a
   
   if (a<dble(1)) then
     if (a<0.01D0) a=0.01D0 ! naive hack when argument is too small
     
     !alternative version:
     !gammadev=rgamma(a+dble(1))*dexp(dlog(rnd())/a)
     
     gammadev=rgamma_alt(a)
   else
     gammadev=rgamma(a)
   endif
  
 end function gammadev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the standard gamma pdf (for a<1)
 double precision function prgamma_alt(a,ujsr)
   double precision :: a,e,x,p,b
   logical :: accept
   integer*8, intent(inout) :: ujsr
   
   prgamma_alt=0.D0
   x=0.D0
   e=dexp(1.D0)
   b=(e+a)/e
   do
     p=prnd(ujsr)*b
     accept=.true.
     if (p>1.D0) then
       x=-dlog((b-p)/a)
       if (dlog(prnd(ujsr))>(a-1.D0)*dlog(x)) accept=.false.
     else
       x=dexp(dlog(p)/a)
       if (x>-dlog(prnd(ujsr))) accept=.false.
     endif
     if (accept) exit
   enddo
   prgamma_alt=x
   
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the standard gamma pdf (for a>1)
 double precision function prgamma(a,ujsr)
   double precision, intent(in) :: a
   double precision :: x,v,u,c,d
   integer*8, intent(inout) :: ujsr
   
   d=a-dble(1)/dble(3)
   c=dble(1)/dsqrt(dble(9)*d)
   do
     do
       x=dsqrt(-dble(2)*dlog(prnd(ujsr)))*dcos(my2pi*prnd(ujsr))
       v=(dble(1)+c*x)
       if (v > dble(0)) exit
     enddo
     v=v**3
     u=prnd(ujsr)
     if (u<(dble(1)-dble(0.0331)*x**4)) exit
     if (dlog(u)<(dble(0.5)*x**2+d*(dble(1)-v+dlog(v)))) exit
   enddo
   prgamma=d*v
   
 end function prgamma
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the standard gamma pdf
 double precision function pgammadev(a,ujsr)
   double precision :: a
   integer*8, intent(inout) :: ujsr
   
   if (a<dble(1)) then
     if (a<0.01D0) a=0.01D0 ! naive hack when argument is too small
     
     !alternative version:
     !gammadev=rgamma(a+dble(1))*dexp(dlog(rnd())/a)
     
     pgammadev=prgamma_alt(a,ujsr)
   else
     pgammadev=prgamma(a,ujsr)
   endif
  
 end function pgammadev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns new value using the rectangular distr. (with reflection)
 double precision function rectdev(x,d)
   double precision, intent(in) :: x,d
   double precision :: nx
   
   nx = x + dble(dble(0.5) - rnd()) * d
   if (nx < dble(0)) then
     nx = -nx
   else
     if (nx > dble(1)) then
       nx = dble(2) - nx
     endif
   endif
   
   rectdev = nx
 end function rectdev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns new value using the rectangular distr. (with reflection)
 double precision function prectdev(x,d,ujsr)
   double precision, intent(in) :: x,d
   double precision :: nx
   integer*8, intent(inout) :: ujsr
   
   nx = x + dble(dble(0.5) - prnd(ujsr)) * d
   if (nx < dble(0)) then
     nx = -nx
   else
     if (nx > dble(1)) then
       nx = dble(2) - nx
     endif
   endif
   
   prectdev = nx
 end function prectdev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the normal pdf (mean m and std.dev. s)
 double precision function normaldev(m,s)
   double precision, intent(in) :: m,s
   double precision :: v1,v2,r
   
   do
     v1 = dble(2) * rnd() - dble(1)
     v2 = dble(2) * rnd() - dble(1)
     r  = v1**2 + v2**2
     if (r < dble(1) .and. r > dble(0)) exit
   enddo
   normaldev = v2 * dsqrt(-dble(2) * dlog(r) / r) * s + m
   
 end function normaldev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine converts degrees to radians
 function deg2rad(theta)
   double precision :: deg2rad,theta
   
   deg2rad=theta/180.D0*mypi
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 function distance(x1,y1,x2,y2)
   double precision :: x1,x2,y1,y2,distance,Erad
   
   ! Average Earth radius used for Lon/Lat coordinates
   Erad=6371.D0
   
   if (planexy) then
     distance=hypot(x1-x2,y1-y2)
   else
     distance=great_circle_distance(x1,y1,x2,y2,Erad)
   endif
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns haversine formula for distance on the Earth surface (Earth approximated as a sphere)
 function great_circle_distance(lon1,lat1,lon2,lat2,rad)
   double precision :: great_circle_distance,lon1,lat1,lon2,lat2,rad
   double precision :: rlon1,rlat1,rlon2,rlat2
   
   rlon1=deg2rad(lon1)
   rlon2=deg2rad(lon2)
   rlat1=deg2rad(lat1)
   rlat2=deg2rad(lat2)
   great_circle_distance=2.D0*rad*dasin(dsqrt(dsin((rlat2-rlat1)*0.5D0)**2+dcos(rlat1)*dcos(rlat2)*dsin((rlon2-rlon1)*0.5D0)**2))
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine sorts real-type vector of length n
 subroutine sort(ra)
   integer :: n
   double precision, intent(inout) :: ra(:)
   integer :: l,j,ir,i
   double precision :: rra
   
   n = size(ra)
   
   l = (n / 2) + 1
   ir = n

   do
     if (l > 1) then
       l = l - 1
       rra = ra(l)
     else
       rra = ra(ir)
       ra(ir) = ra(1)
       ir = ir - 1
       if (ir == 1) then
         ra(1) = rra
         exit
       endif
     endif
   
     i = l
     j = 2 * l
     do while (j <= ir)
       if (j < ir) then
         if (ra(j) < ra(j+1)) then
           j = j + 1
         endif
       endif
     
       if (rra < ra(j)) then
         ra(i) = ra(j)
         i = j
         j = 2 * j
       else
         j = ir + 1 
       endif
     enddo
     ra(i) = rra
   enddo
 end subroutine sort
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes limits of the highest posterior density interval as the shortest range containing
 ! (1-p)*n elements (x is a vector of n double-precision elements) (xl, xu - the output limits)
 subroutine hdi(x,n,p,xl,xu)
   integer, intent(in) :: n
   double precision, intent(inout) :: x(n)
   double precision, intent(in) :: p
   double precision, intent(inout) :: xl,xu
   integer :: i,maxi,r
   double precision :: dul
   
   maxi=int(n*(1.D0-p))
   r=int(n*p)
   call sort(x)
   xu=x(r)
   xl=x(1)
   dul=xu-xl
   do i=1,maxi
     if (x(i+r-1)-x(i)<dul) then
       xu=x(i+r-1)
       xl=x(i)
       dul=xu-xl
     endif
   enddo
   
 end subroutine hdi
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes a statistical support for the model  WAIC (lppd)
 subroutine compute_lppd(ns)
   integer :: ns,i,j
   double precision :: vi,mi, m_waic, waic_i(nind), p_i(nind)
   character (len=255) :: extension
   
   lppd=0.D0
   pw=0.D0
   m_waic=0.D0
   do i=1,nind
     mi=0.D0
     vi=0.D0
     do j=1,ns
       mi=mi+liprob(i,j)/dble(ns)
       vi=vi+dlog(liprob(i,j))/dble(ns)
     enddo
     lppd=lppd+dlog(mi)
     !pw=pw+2.D0*(dlog(mi)-vi) !alternative definition of pWAIC (the default expression for pw needs to be commented out)
     waic_i(i)=dlog(mi)
     mi=vi
     vi=0.D0
     do j=1,ns
       vi=vi+(dlog(liprob(i,j))-mi)**2/dble(ns-1)
     enddo
     waic_i(i)=-2.D0*(waic_i(i)-vi)
     p_i(i)=vi
     m_waic=m_waic+waic_i(i)/dble(nind)
     pw=pw+vi !use this definition by default
   enddo
   
   s_waic=0.D0
   do i=1,nind
     s_waic=s_waic+(waic_i(i)-m_waic)**2/dble(nind-1)
   enddo
   s_waic=dsqrt(dble(nind)*s_waic)
   
   extension='waic'
   output_file=output_filename(extension)
   open (unit=678,file=trim(output_file),action="write",status="replace")
   write (678,"(7a)") 'ind',tab,'pop',tab,'WAIC_i',tab,'Penalty_i'
   do i=1,nind
     write (678,"('ind_',i0,a,'pop_',i0,2(a,e16.8))") inam(i),tab,ipop(i),tab,waic_i(i),tab,p_i(i)
   enddo
   close(678)
 end subroutine compute_lppd
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes mean dispersal distance for a scale parameter (b) and a treshold distance (t)
 ! t - the distance beyond which migration is assumed to be virtually null, t must be provided to get
 ! finite mean distance (NOT USED)
 function disp_kernel_mean(t,b)
   double precision :: tp1,t,b,disp_kernel_mean
   
   tp1=1.D0+t
   if (b==1.D0) then
     disp_kernel_mean=(2.D0*dlog(tp1)-t*(tp1+1.D0))/(2.D0*(t-dlog(tp1)))
   else if (b==2.D0) then
     disp_kernel_mean=(2.D0*tp1*dlog(tp1)-t*(tp1+1.D0))/(t-tp1*dlog(tp1))
   else if (b==3.D0) then
     disp_kernel_mean=(2.D0*tp1*tp1*dlog(tp1)-t*(3.D0*t+2))/(tp1*tp1)
   else
     disp_kernel_mean=(2.D0*tp1**b-tp1*(b*b*t*t+b*t*(2.D0-3.D0*t)+2.D0*(t*t-t+1.D0)))/((b-3.D0)*(tp1**b-tp1*(b*t-t+1.D0)))
   endif
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine writes minimal gnuplot scripts for plotting results
 subroutine write_gnuplot_script
   integer :: out_file,i,k
   character(LEN=255) :: plot_file,extension
   character(len=9), dimension (80) :: hexcod
   
   ! rgb color codes in gnuplot
   hexcod(1)='"#9f0162"'     ! jazzberry jam
   hexcod(2)='"#009f81"'     ! jeepers creepers
   hexcod(3)='"#ff5aaf"'     ! barbie pink
   hexcod(4)='"#00fccf"'     ! aquamarine
   hexcod(5)='"#8400cd"'     ! french violet
   hexcod(6)='"#008df9"'     ! dodger blue
   hexcod(7)='"#ffb2fd"'     ! plum
   hexcod(8)='"#e20134"'     ! alisarin crimson
   hexcod(9)='"#ffc33b"'     ! bright spark
   hexcod(10)='"#003c86"'    ! congress blue
   hexcod(11)='"#804014"'    ! sienna4
   hexcod(12)='"#87ceeb"'    ! skyblue
   hexcod(13)='"#8060ff"'    ! slateblue1
   hexcod(14)='"#a0b6cd"'    ! slategray
   hexcod(15)='"#00ff7f"'    ! spring-green
   hexcod(16)='"#306080"'    ! steelblue
   hexcod(17)='"#ffa040"'    ! tan1
   hexcod(18)='"#40e0d0"'    ! turquoise
   hexcod(19)='"#ee82ee"'    ! violet
   hexcod(20)='"#0080ff"'    ! web-blue
   hexcod(21)='"#00c000"'    ! web-green
   hexcod(22)='"#808000"'    ! yellow4
   hexcod(23)='"#ffff00"'    ! yellow
   hexcod(24)='"#bebebe"'    ! gray
   hexcod(25)='"#00ff00"'    ! green
   hexcod(26)='"#a0ff20"'    ! greenyellow
   hexcod(27)='"#f0fff0"'    ! honeydew
   hexcod(28)='"#ffff80"'    ! khaki1
   hexcod(29)='"#f0e68c"'    ! khaki
   hexcod(30)='"#ffffc0"'    ! lemonchiffon
   hexcod(31)='"#add8e6"'    ! light-blue
   hexcod(32)='"#f08080"'    ! light-coral
   hexcod(33)='"#e0ffff"'    ! light-cyan
   hexcod(34)='"#eedd82"'    ! light-goldenrod
   hexcod(35)='"#d3d3d3"'    ! light-gray
   hexcod(36)='"#90ee90"'    ! light-green
   hexcod(37)='"#f055f0"'    ! light-magenta
   hexcod(38)='"#ffb6c1"'    ! light-pink
   hexcod(39)='"#f03232"'    ! light-red
   hexcod(40)='"#ffa070"'    ! light-salmon
   hexcod(41)='"#afeeee"'    ! light-turquoise
   hexcod(42)='"#ff00ff"'    ! magenta
   hexcod(43)='"#0000cd"'    ! medium-blue
   hexcod(44)='"#8060c0"'    ! mediumpurple3
   hexcod(45)='"#191970"'    ! midnight-blue
   hexcod(46)='"#000080"'    ! navy
   hexcod(47)='"#a08020"'    ! olive
   hexcod(48)='"#ffa500"'    ! orange
   hexcod(49)='"#801400"'    ! orangered4
   hexcod(50)='"#ff4500"'    ! orange-red
   hexcod(51)='"#804080"'    ! orchid4
   hexcod(52)='"#ff80ff"'    ! orchid
   hexcod(53)='"#cdc0b0"'    ! antiquewhite
   hexcod(54)='"#7fffd4"'    ! aquamarine
   hexcod(55)='"#f5f5dc"'    ! beige
   hexcod(56)='"#cdb79e"'    ! bisque
   hexcod(57)='"#0000ff"'    ! blue
   hexcod(58)='"#801414"'    ! brown4
   hexcod(59)='"#a52a2a"'    ! brown
   hexcod(60)='"#7cff40"'    ! chartreuse
   hexcod(61)='"#ff7f50"'    ! coral
   hexcod(62)='"#00ffff"'    ! cyan
   hexcod(63)='"#00008b"'    ! dark-blue
   hexcod(64)='"#408000"'    ! dark-chartreuse
   hexcod(65)='"#00eeee"'    ! dark-cyan
   hexcod(66)='"#b8860b"'    ! dark-goldenrod
   hexcod(67)='"#a0a0a0"'    ! dark-gray
   hexcod(68)='"#006400"'    ! dark-green
   hexcod(69)='"#bdb76b"'    ! dark-khaki
   hexcod(70)='"#c000ff"'    ! dark-magenta
   hexcod(71)='"#556b2f"'    ! dark-olivegreen
   hexcod(72)='"#c04000"'    ! dark-orange
   hexcod(73)='"#ff1493"'    ! dark-pink
   hexcod(74)='"#905040"'    ! dark-plum
   hexcod(75)='"#8b0000"'    ! dark-red
   hexcod(76)='"#e9967a"'    ! dark-salmon
   hexcod(77)='"#008040"'    ! dark-spring-green
   hexcod(78)='"#00ced1"'    ! dark-turquoise
   hexcod(79)='"#9400d3"'    ! dark-violet
   hexcod(80)='"#c8c800"'    ! dark-yellow
   
   
   extension='plt'
   plot_file=output_filename(extension)
   extension='mcmc'
   output_file=output_filename(extension)
   out_file=345
   open (unit=out_file,file=trim(plot_file),action="write",status="replace")
   
   ! general settings
   write (out_file,"(a)") '# general settings'
   write (out_file,"(a)") 'set datafile separator "\t"'
   write (out_file,"(a)") 'set terminal svg enhanced size 350,300'
   write (out_file,"(a)") 'set grid'
   write (out_file,"(a)") 'unset key'
   write (out_file,"(a)") 'set tics font ",15"'
   write (out_file,"(a)") 'set xlabel font ",15"'
   write (out_file,"(a)") 'set ylabel font ",15"'
   write (out_file,"(a)") 'set lmargin 10'
   write (out_file,"(a)") 'set bmargin 3'
   
   ! isolation parameters
   if (zygotic_migration) then
     k=3
     k=k+npop**2+2
     
     ! kernel density plot for z_alpha
     write (out_file,"(a)") ''
     write (out_file,"(a)") '# density plot for z_alpha'
     write (out_file,"(a)") 'set output "z_alpha_dens.svg"' 
     write (out_file,"(a)") 'set xlabel "z_{/Symbol a}"'
     write (out_file,"(a)") 'set ylabel "Density"'
     write (out_file,"(a,i0,a,i0,a)") 'plot "'//trim(output_file)//'" u ',k,':(1./',ns/ke,'.) smooth kdensity lt 2 lw 2'
   endif
   
   if (zygotic_migration .and. spatial_zygotic) then
     k=3+npop**2+3
     
     ! kernel density plot for b_alpha
     write (out_file,"(a)") ''
     write (out_file,"(a)") '# density plot for b_alpha'     
     write (out_file,"(a)") 'set output "b_alpha_dens.svg"' 
     write (out_file,"(a)") 'set xlabel "b_{/Symbol a}"'
     write (out_file,"(a)") 'set ylabel "Density"'
     write (out_file,"(a,i0,a,i0,a)") 'plot "'//trim(output_file)//'" u ',k,':(1./',ns/ke,'.) smooth kdensity lt 2 lw 2'
   endif
   
   if (gametic_migration) then
     k=3
     if (zygotic_migration) k=k+npop**2+2
     if (zygotic_migration .and. spatial_gametic) k=k+1
     k=k+npop**2+2
     
     ! kernel density plot for z_beta
     write (out_file,"(a)") ''
     write (out_file,"(a)") '# density plot for z_beta'     
     write (out_file,"(a)") 'set output "z_beta_dens.svg"' 
     write (out_file,"(a)") 'set xlabel "z_{/Symbol b}"'
     write (out_file,"(a)") 'set ylabel "Density"'
     write (out_file,"(a,i0,a,i0,a)") 'plot "'//trim(output_file)//'" u ',k,':(1./',ns/ke,'.) smooth kdensity lt 2 lw 2'
   endif
   
   if (gametic_migration .and. spatial_gametic) then
     k=3
     if (zygotic_migration) k=k+npop**2+2
     if (zygotic_migration .and. spatial_gametic) k=k+1
     k=k+npop**2+3
     
     ! kernel density plot for b_beta
     write (out_file,"(a)") ''
     write (out_file,"(a)") '# density plot for b_beta'
     write (out_file,"(a)") 'set output "b_beta_dens.svg"' 
     write (out_file,"(a)") 'set xlabel "b_{/Symbol b}"'
     write (out_file,"(a)") 'set ylabel "Density"'
     write (out_file,"(a,i0,a,i0,a)") 'plot "'//trim(output_file)//'" u ',k,':(1./',ns/ke,'.) smooth kdensity lt 2 lw 2'
   endif
   
   if (dropout) then
     k=3
     if (zygotic_migration) k=k+npop**2+2
     if (zygotic_migration .and. spatial_gametic) k=k+1
     if (gametic_migration) k=k+npop**2+2
     if (gametic_migration .and. spatial_gametic) k=k+1
     write (out_file,"(a)") ''
     write (out_file,"(a)") '# boxplot for allelic dropout' 
     write (out_file,"(a)") 'set output "dropout_box.svg"'
     write (out_file,"(a)") 'set style boxplot nooutliers' 
     write (out_file,"(a)") 'set style data boxplot'
     write (out_file,"(a)") 'set style fill transparent solid 0.7 noborder' 
     write (out_file,"(a)") 'set errorbars 0.5' 
     write (out_file,"(a)") 'set xlabel "Locus"'
     write (out_file,"(a)") 'set ylabel "Allelic dropout"'
     write (out_file,"(a)") 'unset key'
     write (out_file,"(a)", advance='no') 'set xtics ('
     do i=1,nloc-1
       write (out_file,"(2(a,i0),',')", advance='no') '"',i,'" ',i
     enddo
     write (out_file,"(2(a,i0),')')", advance='yes') '"',nloc,'" ',nloc
     do i=1,nloc-1
       k=k+1
       if (i==1) then
         write (out_file,"(a,2(i0,a))") 'plot "'//trim(output_file)//'" u (',i,'):',k,' lc '//hexcod(2)//', \'
       else
         write (out_file,"(2(a,i0),a)") '"'//trim(output_file)//'" u (',i,'):',k,' lc '//hexcod(2)//', \'
       endif
     enddo
     k=k+1
     write (out_file,"(2(a,i0),a)") '"'//trim(output_file)//'" u (',nloc,'):',k,' lc '//hexcod(2)
     write (out_file,"(a)") 'set xtics auto'
   endif
   
   if (inbreeding)  then
     k=3
     if (zygotic_migration) k=k+npop**2+2
     if (zygotic_migration .and. spatial_gametic) k=k+1
     if (gametic_migration) k=k+npop**2+2
     if (gametic_migration .and. spatial_gametic) k=k+1
     if (dropout) k=k+nloc
     k=k+nind+1
     
     ! kernel density plot for m_r
     write (out_file,"(a)") ''
     write (out_file,"(a)") '# density plot for mean_inbreeding'        
     write (out_file,"(a)") 'set output "mean_inbreeding_dens.svg"' 
     write (out_file,"(a)") 'set xlabel "Mean inbreeding"'
     write (out_file,"(a)") 'set ylabel "Density"'
     write (out_file,"(a,i0,a,i0,a)") 'plot "'//trim(output_file)//'" u ',k,':(1./',ns/ke,'.) smooth kdensity lt 2 lw 2'
   endif
   
   ! kernel density plot for fst
   k=3
   if (zygotic_migration) k=k+npop**2+2
   if (zygotic_migration .and. spatial_gametic) k=k+1
   if (gametic_migration) k=k+npop**2+2
   if (gametic_migration .and. spatial_gametic) k=k+1
   if (dropout) k=k+nloc
   if (inbreeding) k=k+nind+2
   write (out_file,"(a)") ''
   write (out_file,"(a)") '# density plot for divergence' 
   do i=1,npop
     k=k+1
     if (i==1) then
       write (out_file,"(a)") 'set output "divergence_dens.svg"' 
       write (out_file,"(a)") 'set xlabel "Divergence"'
       write (out_file,"(a)") 'set ylabel "Density"'
       write (out_file,"(a)") 'set key'
       write (out_file,"(a,3(i0,a))") 'plot "'//trim(output_file)//'" u ',k,':(1./',ns/ke,'.) smooth kdensity lt 2 lw 1 &
       & lc '//hexcod(i)//' title "Pop ',pop(i),'", \' 
     else
       write (out_file,"(a,3(i0,a))") '"'//trim(output_file)//'" u ',k,':(1./',ns/ke,'.) smooth kdensity lw 1 &
       & lc '//hexcod(i)//' title "Pop ',pop(i),'", \'
     endif
   enddo
   
   k=k+1
   ! kernel density plot for m_fst
   write (out_file,"(a,i0,a,i0,a)") '"'//trim(output_file)//'" u ',k,':(1./',ns/ke,'.) smooth kdensity lw 3 title "Mean"'
   
   ! boxplot for fst
   k=3
   if (zygotic_migration) k=k+npop**2+2
   if (zygotic_migration .and. spatial_gametic) k=k+1
   if (gametic_migration) k=k+npop**2+2
   if (gametic_migration .and. spatial_gametic) k=k+1
   if (dropout) k=k+nloc
   if (inbreeding) k=k+nind+2
   write (out_file,"(a)") ''
   write (out_file,"(a)") '# boxplot for divergence'
   write (out_file,"(a)") 'set style boxplot nooutliers' 
   write (out_file,"(a)") 'set style data boxplot'
   write (out_file,"(a)") 'set style fill transparent solid 0.7 noborder' 
   write (out_file,"(a)") 'set errorbars 0.5' 
   write (out_file,"(a)") 'set output "divergence_box.svg"' 
   write (out_file,"(a)") 'set xlabel "Population"'
   write (out_file,"(a)") 'set ylabel "Divergence"'
   write (out_file,"(a)") 'unset key'
   write (out_file,"(a)", advance='no') 'set xtics ('
   do i=1,npop-1
     write (out_file,"(2(a,i0),',')", advance='no') '"',pop(i),'" ',i
   enddo
   write (out_file,"(2(a,i0),')')", advance='yes') '"',pop(npop),'" ',npop
   
   do i=1,npop-1
     k=k+1
     if (i==1) then
       write (out_file,"(a,2(i0,a))") 'plot "'//trim(output_file)//'" u (',i,'):',k,' lc '//hexcod(i)//', \'
     else
       write (out_file,"(2(a,i0),a)") '"'//trim(output_file)//'" u (',i,'):',k,' lc '//hexcod(i)//', \'
     endif
   enddo
   k=k+1
   write (out_file,"(2(a,i0),a)") '"'//trim(output_file)//'" u (',npop,'):',k,' lc '//hexcod(i)
   write (out_file,"(a)") 'set xtics auto'
   
   ! scatter plots for alpha and beta (if coordinates are available)
   if (zygotic_migration .and. (coordinates .or. planexy)) then
     write (out_file,"(a)") ''
     write (out_file,"(a)") '# scatter plot for alpha'
     do i=1,npop
       write (out_file,"('$alpha',i0,' << EOD')") i
       do k=1,npop
         if (i==k) cycle
         write (out_file,"(e14.6,a,f7.5)") pdist(i,k),tab,malpha(i,k)
       enddo
       write (out_file,"(a)") 'EOD'
     enddo
     write (out_file,"(a)") 'set ylabel "Migration ({/Symbol a})"'
     write (out_file,"(a)") 'set xlabel "Distance"'
     write (out_file,"(a)") 'set logscale y'
     write (out_file,"(a)") 'set format y "%g"'
     write (out_file,"(a)") 'set output "distalpha.svg"'
     write (out_file,"(a)") 'plot $alpha1 u 1:2 w p pt 7 lc '//hexcod(1)//', \'
     do i=2,npop-1
       write (out_file,"(a,i0,a)") '$alpha',i,' u 1:2 w p pt 7 lc '//hexcod(i)//', \'
     enddo
     write (out_file,"(a,i0,a)") '$alpha',npop,' u 1:2 w p pt 7 lc '//hexcod(npop)
   endif
   
   if (gametic_migration .and. (coordinates .or. planexy)) then
     write (out_file,"(a)") ''
     write (out_file,"(a)") '# scatter plot for beta'
     do i=1,npop
       write (out_file,"('$beta',i0,' << EOD')") i
       do k=1,npop
         if (i==k) cycle
         write (out_file,"(e14.6,a,f7.5)") pdist(i,k),tab,mbeta(i,k)
       enddo
       write (out_file,"(a)") 'EOD'
     enddo
     write (out_file,"(a)") 'set ylabel "Migration ({/Symbol b})"'
     write (out_file,"(a)") 'set xlabel "Distance"'
     write (out_file,"(a)") 'set logscale y'
     write (out_file,"(a)") 'set output "distbeta.svg"'
     write (out_file,"(a)") 'plot $beta1 u 1:2 w p pt 7 lc '//hexcod(1)//', \'
     do i=2,npop-1
       write (out_file,"(a,i0,a)") '$beta',i,' u 1:2 w p pt 7 lc '//hexcod(i)//', \'
     enddo
     write (out_file,"(a,i0,a)") '$beta',npop,' u 1:2 w p pt 7 lc '//hexcod(npop)
   endif
   
   ! stacked bar chart for alpha
   write (out_file,"(a)") ''
   write (out_file,"(a)") '# bar chart for alpha'
   write (out_file,"(a)") '$alpha << EOD'
   do i=1,npop
     write (out_file,"(*(f7.5,' '))") (malpha(i,k),k=1,npop-1),malpha(i,npop)
   enddo
   write (out_file,"(a)") 'EOD'
   write (out_file,"(a)") 'reset'
   write (out_file,"(a)") 'unset key'
   write (out_file,"(a)") 'set grid'
   write (out_file,"(a)") 'set tics font ",15"'
   write (out_file,"(a)") 'set xlabel font ",15"'
   write (out_file,"(a)") 'set ylabel font ",15"'
   write (out_file,"(a)") 'set lmargin 10'
   write (out_file,"(a)") 'set bmargin 3'
   write (out_file,"(a)") 'set style data histograms'
   write (out_file,"(a)") 'set style histogram rowstacked'
   write (out_file,"(a)") 'set style fill transparent solid 0.8 noborder'
   write (out_file,"(a)") 'set yrange [0:1]'
   write (out_file,"(a)") 'set ylabel "Migration ({/Symbol a})"'
   write (out_file,"(a)") 'set xlabel "Population"'
   write (out_file,"(a)") 'set output "alpha.svg"'
   write (out_file,"(a)", advance='no') 'set xtics ('
   do i=1,npop-1
     write (out_file,"(2(a,i0),',')", advance='no') '"',pop(i),'" ',i-1
   enddo
   write (out_file,"(2(a,i0),')')", advance='yes') '"',pop(npop),'" ',npop-1
   
   write (out_file,"(a)",advance="no") 'plot $alpha u 1 lc '//hexcod(1)
   write (out_file,"(*(a,i0,a))",advance="yes") (', "" u ',i,' lc '//hexcod(i),i=2,npop)
   
   ! stacked bar chart for beta
   write (out_file,"(a)") ''
   write (out_file,"(a)") '# bar chart for beta'
   write (out_file,"(a)") '$beta << EOD'
   do i=1,npop
     write (out_file,"(*(f7.5,' '))") (mbeta(i,k),k=1,npop-1),mbeta(i,npop)
   enddo
   write (out_file,"(a)") 'EOD'
   write (out_file,"(a)") 'reset'
   write (out_file,"(a)") 'unset key'
   write (out_file,"(a)") 'set grid'
   write (out_file,"(a)") 'set tics font ",15"'
   write (out_file,"(a)") 'set xlabel font ",15"'
   write (out_file,"(a)") 'set ylabel font ",15"'
   write (out_file,"(a)") 'set lmargin 10'
   write (out_file,"(a)") 'set bmargin 3'
   write (out_file,"(a)") 'set style data histograms'
   write (out_file,"(a)") 'set style histogram rowstacked'
   write (out_file,"(a)") 'set style fill transparent solid 0.8 noborder'
   write (out_file,"(a)") 'set yrange [0:1]'
   write (out_file,"(a)") 'set ylabel "Migration ({/Symbol b})"'
   write (out_file,"(a)") 'set xlabel "Population"'
   write (out_file,"(a)") 'set output "beta.svg"'
   write (out_file,"(a)", advance='no') 'set xtics ('
   do i=1,npop-1
     write (out_file,"(2(a,i0),',')", advance='no') '"',pop(i),'" ',i-1
   enddo
   write (out_file,"(2(a,i0),')')", advance='yes') '"',pop(npop),'" ',npop-1
   
   write (out_file,"(a)",advance="no") 'plot $beta u 1 lc '//hexcod(1)
   write (out_file,"(*(a,i0,a))",advance="yes") (', "" u ',i,' lc '//hexcod(i),i=2,npop)
   
   close (out_file)
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine saves data for BayesAss
 subroutine BA3_data_file
   integer :: k,l,m,tmp_pop,tmp_gen(nloc,2), gen(2)
   character(LEN=255) :: BA3_file,extension
   
   extension='BA3'
   BA3_file=output_filename(extension)
   
   ! sort data to meet BayesAss requirements
   do k=1,nind-1
     do l=k+1,nind
       if (ipop(k)>ipop(l)) then
         tmp_pop=ipop(k)
         tmp_gen=igen(k,:,:)
         ipop(k)=ipop(l)
         igen(k,:,:)=igen(l,:,:)
         ipop(l)=tmp_pop
         igen(l,:,:)=tmp_gen
       endif
     enddo
   enddo
   
   open (unit=789,file=BA3_file,action="write",status="replace")
   do l=1,nloc
     do m=1,nind
       gen=igen(m,l,:)
       if (gen(1)<0) gen(1)=0
       if (gen(2)<0) gen(2)=0
       write (789,"('ind',i0,' pop',i0,' loc',i0,2(' ',i0))") m,ipop(m),l,gen(1),gen(2)
     enddo
   enddo
   close (789)
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine saves data for BayesAss
 subroutine BIMr_data_file
   integer :: i,l,tmp_gen(nloc,2)
   character(LEN=255) :: BIMr_file,extension
   
   extension='gen'
   BIMr_file=output_filename(extension)
   
   open (unit=789,file=BIMr_file,action="write",status="replace")
   
   write (789,"('[individuals]= ',i0)") nind
   write (789,"('[populations]= ',i0)") npop
   write (789,"('[loci]= ',i0)") nloc
   write (789,"('[alleles]= ',i0)") maxnall
   
   write (789,"('[genotypes]= ',i0)")
   do i=1,nind
     tmp_gen=igen(i,:,:)
     do l=1,nloc
       if (tmp_gen(l,1)<0) tmp_gen(l,1)=0
       if (tmp_gen(l,2)<0) tmp_gen(l,2)=0
     enddo
     
     write (789,"('ind',i0,' ,',a,i0,*(a,2i2.2))") inam(i),tab,ipop(i),(tab,tmp_gen(l,1),tmp_gen(l,2),l=1,nloc)
   enddo
   close (789)
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
end program
