!  rspmisim.f90
!  
!  Copyright 2024 Igor J. Chybicki <igorchy@ukw.edu.pl>
!  
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
!  
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!  
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
!  MA 02110-1301, USA.
!  
!  INFO:
!  The program generates random genotype data under a simulation
!  model that resembles the inference model implemented in rspmi.
!  
!  As the program has no interface, parameter settings can be only changed
!  before compiling. See the subsection 'simulation settings' in the main
!  program below.
!  
!  The program was tested with the GNU Fortran (GFortran) compiler only.
!  The minimum command to compile the code is:
!  gfortran -o rspmisim rspmisim.f90
!  As a result, an executable binary file (rspmisim) is created.

program spmisim
   
   implicit none
   
   ! ----------------------
   ! global declarations
   !
   ! useful constants
   character(1), parameter :: tab=char(9)
   double precision, parameter :: mypi = dacos(dble(-1.0)), my2pi = 2.D0 * mypi
   
   ! useful variables
   integer, dimension(8) :: values
   character(256) :: title
   
   ! RNG-related variables
   integer*8 :: seed,jsr
   
   ! data dimensions
   integer :: npop, nind, nloc, nall
   
   ! simulation replicability
   integer :: repl, nrepl
   
   ! simulation hyper-parameters
   double precision :: y_alpha,y_beta, z_alpha, z_beta, b_alpha, b_beta, m_f, y_f, min_f, max_f, m_r, y_r, pr0
   
   ! simulation parameters
   double precision, allocatable :: qlm(:,:), pklm(:,:,:), fk(:), alpha(:,:), beta(:,:), ri(:), r_alpha(:,:), r_beta(:,:)
   logical :: constant_divergence=.false.
   logical :: coordinates=.false.
   
   ! simulated data
   integer, allocatable :: ipop(:), igen(:,:,:), ianc(:,:)
   double precision, allocatable :: ix(:), iy(:), pdist(:,:)
   
   ! ----------------------
   ! program starts here
   ! ----------------------
   
   ! rng seeded with the time (in milliseconds) since midnight
   call date_and_time(values=values)
   seed=((values(5)*60+values(6))*60+values(7))*1000+values(8)
   jsr=seed
   call init_rnd
   
   ! ----------------------
   ! simulation settings
   nrepl = 10            ! number of replications
   
   ! data dimensions
   npop = 10             ! number of populations
   nind = 1000           ! total number of individuals
   nloc = 20             ! number of markers
   nall = 6              ! number of alleles per locus
   coordinates = .false. ! if TRUE, coordinates are attached to populations
   
   ! hyper-parameters
   y_alpha = 0.1D0            ! dispersion of seed migration rates around their expectations
   y_beta  = 0.1D0            ! dispersion of pollen migration rates around their expectations
   
   z_alpha = 0.25D0           ! isolation factor for seed migration
   z_beta  = 0.25D0           ! isolation factor for pollen migration
   
   !b_alpha = 2.127354216D0    ! seed dispersal kernel (strong effect)
   !b_alpha = 1.206181141D0    ! seed dispersal kernel (medium effect)
   b_alpha = 0.D0             ! seed dispersal kernel (no effect)
   
   !b_beta = 2.127354216D0     ! pollen dispersal kernel (strong effect)
   !b_beta  = 1.206181141D0    ! pollen dispersal kernel (medium effect)
   b_beta  = 0.D0             ! pollen dispersal kernel (no effect)
   
   m_f = 0.20D0               ! mean divergence rate
   y_f = 0.01D0               ! dispersion of divergence rates among pops (Beta parameter alpha=m_k*(1-y_f)/(y_f))
   min_f=0.01D0               ! minimum divergence value allowed
   max_f=0.9D0                ! maximum divergence value allowed
   
   m_r = 0.D0                 ! mean inbreeding
   y_r = 0.1D0                ! dispersion of individual inbreeding
   pr0 = 0.D0                 ! zero-inflated beta distribution parameter (simple beta distrib. if pr0=0) (NOT USED)
   
   if (y_f==0.D0) constant_divergence=.true.
   ! ----------------------
   
   do repl=1,nrepl
     write (title,"('repl',i0)") repl
     call simulate_data
     write (*,*) 'Simulated data saved to "', trim(title),'"'
   enddo
   
   ! ----------------------
   ! program ends here
   ! ----------------------
   
 contains
 
 !----------------------------------------------------------------------------------------------------------
 ! sorts data per pop (required for proper BayesAss data reading)
 subroutine sort_ind_per_pop
   integer :: i,j,tmp_pop,tmp_gen(nloc,2),tmp_anc(2)
   double precision :: tmp_r
   
   do i=1,nind-1
     do j=i+1,nind
       if (ipop(i)>ipop(j)) then
         tmp_pop=ipop(i)
         tmp_gen=igen(i,:,:)
         ipop(i)=ipop(j)
         igen(i,:,:)=igen(j,:,:)
         ipop(j)=tmp_pop
         igen(j,:,:)=tmp_gen
         tmp_r=ri(i)
         ri(i)=ri(j)
         ri(j)=tmp_r
         tmp_anc=ianc(i,:)
         ianc(i,:)=ianc(j,:)
         ianc(j,:)=tmp_anc
       endif
     enddo
   enddo
 end subroutine sort_ind_per_pop
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! simulates genetic data based on the zygotic/gametic migration model
 subroutine simulate_data
   integer :: k,l,m,n
   double precision :: sa,sb
   double precision, allocatable :: ak(:)
   character(len=256) :: file_name
   
   ! spatial data requires exactly 9 populations
   if (coordinates .and. npop/=9) then
     write (*,*) 'For a spatial model, the number of populations is set to 9'
     npop = 9
   endif
   
   ! memory allocations
   allocate(qlm(nloc,nall))
   allocate(pklm(npop,nloc,nall))
   allocate(fk(npop))
   allocate(alpha(npop,npop))
   allocate(beta(npop,npop))
   allocate(ipop(nind))
   allocate(igen(nind,nloc,2))
   allocate(ianc(nind,2))
   allocate(ak(max(npop,nall,2)))
   allocate(ri(nind))
   allocate(ix(npop))
   allocate(iy(npop))
   allocate(r_alpha(npop,npop))
   allocate(r_beta(npop,npop))
   allocate(pdist(npop,npop))
   
   if (coordinates) then
     ! spatial data (coordinates replicate the distribution of Taxus populations, in km)
     ix(1)= 5.97248284419462;   iy(1)=-2.95656791140371
     ix(2)=-0.07602457865386;   iy(2)= 1.28117428530386
     ix(3)=-0.82431975807321;   iy(3)= 1.68099369443372
     ix(4)=-2.24911701169324;   iy(4)= 2.14446311907702
     ix(5)= 9.45762704323155;   iy(5)= 0.00022111564227
     ix(6)=-3.61961678859432;   iy(6)= 0.86194397883321
     ix(7)=-4.18229602968304;   iy(7)= 1.86278109905912
     ix(8)= 5.69439994842779;   iy(8)=-5.55599489127804
     ix(9)=-10.1731353846096;   iy(9)= 0.68098586687799
   else
     ix=0.D0; iy=0.D0
   endif
   
   ! distances between populations
   do k=1,npop
     do l=1,npop
       pdist(k,l)=hypot(ix(k)-ix(l),iy(k)-iy(l))
     enddo
   enddo
   
   ! expected values of migration dispersal kernel for seeds and pollen
   do k=1,npop
     r_alpha(k,k)=0.D0
     r_beta(k,k)=0.D0
     sa=r_alpha(k,k)
     sb=r_beta(k,k)
     do l=1,npop
       if (l==k) cycle
       r_alpha(k,l)=dexp(-b_alpha*dlog(1.D0+pdist(k,l)))
       r_beta(k,l)=dexp(-b_beta*dlog(1.D0+pdist(k,l)))
       sa=sa+r_alpha(k,l)
       sb=sb+r_beta(k,l)
     enddo
     r_alpha(k,:)=r_alpha(k,:)/sa
     r_beta(k,:)=r_beta(k,:)/sb
   enddo
   
   ! uniform prior global allele frequencies (for the F-model)
   qlm=1.D0/dble(nall)
   
   ! divergence rates (for the F-model)
   do k=1,npop
     if (constant_divergence) then
       fk(k)=m_f
     else
       do
         sa=0.D0 
         fk(k)=gammadev(m_f * (1.D0-y_f)/y_f)
         sa=fk(k)+gammadev((1.D0-m_f) * (1.D0-y_f)/y_f)
         fk(k)=fk(k)/sa
         if (fk(k)>min_f .and. fk(k)<max_f) exit
       enddo
     endif
     
     ! expected migration rates
     ! the prior isolation (non-migration) is set to 2/3
     do l=1,npop
       alpha(k,l)=(1.D0-z_alpha)/3.D0*r_alpha(k,l)
       beta(k,l)=(1.D0-z_beta)/3.D0*r_beta(k,l)
     enddo
     alpha(k,k)=(2.D0+z_alpha)/3.D0
     beta(k,k) =(2.D0+z_beta)/3.D0
     
     ! random zygotic migration rates taken as draws from Dirichlet distrib.
     do
       sa=0.D0
       do l=1,npop
         ak(l)=gammadev(alpha(k,l) * (1.D0-y_alpha)/y_alpha)
         sa=sa+ak(l)
       enddo
       alpha(k,:)=ak(1:npop)/sa
       
       ! to assure alpha(k,k) is the maximum element in a vector
       if (maxval(alpha(k,:))==alpha(k,k)) exit
     enddo
     
     ! random gametic migration rates taken as draws from Dirichlet distrib.
     do
       sa=0.D0
       do l=1,npop
         ak(l)=gammadev(beta(k,l) * (1.D0-y_beta)/y_beta)
         sa=sa+ak(l)
       enddo
       beta(k,:)=ak(1:npop)/sa
       
       ! to assure beta(k,k) is the maximum element in a vector
       if (maxval(beta(k,:))==beta(k,k)) exit
     enddo
     
     ! random allele frequencies in pops taken as draws from Dirichlet distrib.
     do l=1,nloc
       sa=0.D0
       do m=1,nall
         ak(m)=gammadev(qlm(l,m) * (1.D0-fk(k))/fk(k))
         sa=sa+ak(m)
       enddo
       pklm(k,l,:)=ak(1:nall)/sa
     enddo
   enddo
   
   ! draws a random sample of individual genotypes
   do k=1,nind
     ! draws sampling location
     ipop(k)=int(rnd()*dble(npop))+1
     
     ! draws seed and pollen origin
     ! double migration possible, as in real populations
     m=draw_pop(alpha(ipop(k),:)) ! random seed origin
     n=draw_pop(beta(m,:))        ! random pollen origin
     ianc(k,1)=m
     ianc(k,2)=n
     
     ! draws inbreeding if both gametes from the same pop
     ! otherwise, inbreeding is zero
     if (m==n .and. m_r>0.D0) then
       if (rnd()<(1.D0-pr0)) then ! if zero-inflated Beta distribution is needed, pr0 should be greater than zero
         if (y_r>0.D0) then
           ak(1)=gammadev(m_r * (1.D0-y_r)/y_r)
           ak(2)=gammadev((1.D0-m_r) * (1.D0-y_r)/y_r)
           ri(k)=ak(1)/(ak(1)+ak(2))
         else
           ri(k)=m_r
         endif
       else
         ri(k)=0.D0
       endif
     else
       ri(k)=0.D0  
     endif
     
     ! generates multilocus genotype based on seed and pollen origin
     do l=1,nloc
       igen(k,l,1)=draw_allele(m,l)
       if (rnd()<ri(k)) then
         igen(k,l,2)=igen(k,l,1)
       else
         igen(k,l,2)=draw_allele(n,l)
       endif
     enddo
   enddo
   
   ! for example, BayesAss (or BA3) expects sorted data according to sampling individuals, locations, and loci
   call sort_ind_per_pop
   
   ! creates BayesAss (BA3) data file
   file_name=trim(title)//'.ba3'
   open (unit=789,file=file_name,action="write",status="replace")
   do l=1,nloc
     do m=1,nind
       write (789,"('ind',i0,' pop',i0,' loc',i0,2(' ',i0))") m,ipop(m),l,igen(m,l,1),igen(m,l,2)
     enddo
   enddo
   close (789)
   
   ! creates rspmi data file
   file_name=trim(title)//'.rspmi'
   open (unit=789,file=file_name,action="write",status="replace")
   if (coordinates) then
     do m=1,nind
       write (789,"(2(i0,a),e16.8,a,e16.8,*(a,i0))") m,tab,ipop(m),tab,ix(ipop(m)),tab,iy(ipop(m)),(tab,igen(m,l,1), &
       tab,igen(m,l,2),l=1,nloc)
     enddo
   else
     do m=1,nind
       write (789,"(i0,a,i0,*(a,i0))") m,tab,ipop(m),(tab,igen(m,l,1), &
       tab,igen(m,l,2),l=1,nloc)
   enddo
   endif
   close (789)
   
   ! creates rspmi ideal migration data file
   file_name=trim(title)//'.ideal'
   open (unit=789,file=file_name,action="write",status="replace")
   if (coordinates) then
     do m=1,nind
       write (789,"(2(i0,a),2(e16.8,a),i0,a,i0)") m,tab,ipop(m),tab,ix(ipop(m)),tab,iy(ipop(m)),tab,ianc(m,1),tab,ianc(m,2)
     enddo
   else
     write (789,"(i0,3(a,i0))") m,tab,ipop(m),tab,ianc(m,1),tab,ianc(m,2)
   endif
   close (789)
   
   ! creates file with expected migration rates
   file_name=trim(title)//'.params'
   open (unit=789,file=file_name,action="write",status="replace")
   write (789,"(a)") 'Expected migration rates:'
   write (789,"('Sink',a,'Source',a,'alpha',a,'beta')") tab,tab,tab
   do k=1,npop
     do l=1,npop
       if (k/=l) write (789,"(2('Pop_',i0,a),f7.4,a,f7.4)") k,tab,l,tab,alpha(k,l),tab,beta(k,l)
     enddo
   enddo
   write (789,"(a)") ''
   write (789,"(a)") 'Expected divergence rates:'
   write (789,"('Pop',a,'fst')") tab
   do k=1,npop
     write (789,"('Pop_',i0,a,f7.4)") k,tab,fk(k)
   enddo
   write (789,"(a)") ''
   write (789,"(a)") 'Expected inbreeding:'
   write (789,"('Ind',a,'fis')") tab
   do k=1,nind
     write (789,"('Ind_',i0,a,f7.4)") k,tab,ri(k)
   enddo
   
   write (789,"(a)") ''
   write (789,"(a)") 'Hyper-parameters:'
   write (789,"(2a,f7.4)") 'y_alpha',tab,y_alpha
   write (789,"(2a,f7.4)") 'y_beta',tab,y_beta
   write (789,"(2a,f7.4)") 'z_alpha',tab,z_alpha
   write (789,"(2a,f7.4)") 'z_beta',tab,z_beta
   write (789,"(2a,f7.4)") 'b_alpha',tab,b_alpha
   write (789,"(2a,f7.4)") 'b_beta',tab,b_beta
   write (789,"(2a,f7.4)") 'm_fst',tab,m_f
   write (789,"(2a,f7.4)") 'm_fis',tab,m_r
   write (789,"(2a,f7.4)") 'y_fis',tab,y_r
   close(789)
   
   ! frees memory
   deallocate(qlm)
   deallocate(pklm)
   deallocate(fk)
   deallocate(alpha)
   deallocate(beta)
   deallocate(ipop)
   deallocate(igen)
   deallocate(ianc)
   deallocate(ak)
   deallocate(ri)
   deallocate(ix)
   deallocate(iy)
   deallocate(r_alpha)
   deallocate(r_beta)
   deallocate(pdist)
   
 end subroutine simulate_data
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! draws a random allele based on allele freqs in the k-th pop and l-th loc
 function draw_allele(pop,loc)
   integer :: draw_allele,m,pop,loc
   double precision :: r,sp
   
   draw_allele=1
   r=rnd()
   sp=0.D0
   do m=1,nall
     sp=sp+pklm(pop,loc,m)
     if (r<sp) then
       draw_allele=m
       return
     endif
   enddo
   
 end function draw_allele
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! draws pop number based on migration vector m
 function draw_pop(m)
   integer :: draw_pop,k
   double precision :: r,m(npop),sm
   
   draw_pop=1
   r=rnd()
   sm=0.D0
   do k=1,npop
     sm=sm+m(k)
     if (r<sm) then
       draw_pop=k
       return
     endif
   enddo
   
 end function draw_pop
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! implements a shift-register generator (64-bit version)
 function shift_xor(val,shift)
   integer*8 :: shift_xor
   integer*8, intent(in) :: val, shift
   
   shift_xor=ieor(val,ishft(val,shift))
 end function shift_xor
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! initializes random number generator
 subroutine init_rnd
    
    jsr=shift_xor(shift_xor(shift_xor(jsr, 13_8),-17_8),43_8)
 end subroutine init_rnd
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! returns a random draw from the uniform (0,1) pdf
 double precision function rnd()
   integer*8 :: ojsr
   ojsr=jsr
   jsr=shift_xor(shift_xor(shift_xor(jsr, 13_8),-17_8),43_8)
   rnd=dble(ojsr+jsr)*5.42101086242752217D-20+0.5D0
 end function rnd
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! returns a random draw from the standard gamma pdf (should be used for a>1)
 double precision function rgamma(a)
   double precision, intent(in) :: a
   double precision :: x,v,u,c,d
   
   d=a-dble(1)/dble(3)
   c=dble(1)/dsqrt(dble(9)*d)
   
   do
     do
       x=dsqrt(-dble(2)*dlog(rnd()))*dcos(my2pi*rnd())
       v=(dble(1)+c*x)
       if (v > dble(0)) exit
     enddo
     
     v=v**3
     u=rnd()
     
     if (u<(dble(1)-dble(0.0331)*x**4)) exit
     if (dlog(u)<(dble(0.5)*x**2+d*(dble(1)-v+dlog(v)))) exit
   enddo
   
   rgamma=d*v
 end function rgamma
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! returns a random draw from the standard gamma pdf
 double precision function gammadev(a)
   double precision :: a
  
   if (a<dble(1)) then
     if (a<0.01D0) a=0.01D0
     gammadev=rgamma(a+dble(1))*rnd()**(dble(1)/a)
   else
     gammadev=rgamma(a)
   endif
  
 end function gammadev
 !----------------------------------------------------------------------------------------------------------
 
end program
